static char rcsid[] = "$Id: stage1hr.c 215540 2018-05-30 21:55:55Z twu $";
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifndef HAVE_MEMCPY
#define memcpy(d,s,n) bcopy((s),(d),(n))
#endif
#ifndef HAVE_MEMMOVE
#define memmove(d,s,n) bcopy((s),(d),(n))
#endif

#include "stage1hr.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>		/* For memset() */
#include <math.h>
#include <ctype.h>		/* for tolower() */
#include "assert.h"
#include "mem.h"
#include "types.h"		/* Needed for HAVE_64_BIT */

#include "reader.h"
#include "oligo.h"
#include "list.h"
#include "intlist.h"
#include "splice.h"
#include "indel.h"
#include "stage3hr.h"
#include "concordance.h"
#include "substring.h"
#include "complement.h"
#include "compress.h"
#include "genome128_hr.h"
#include "genome_sites.h"
#include "maxent.h"
#include "maxent_hr.h"
#include "iitdef.h"
#include "univinterval.h"
#ifdef LARGE_GENOMES
#include "uint8list.h"
#else
#include "uintlist.h"
#endif
#include "record.h"

#ifdef HAVE_64_BIT
/* Oligospace_T is 64 bits */
#include "uint8table.h"
#else
/* Oligospace_T is 32 bits */
#include "uinttable.h"
#endif


#include "orderstat.h"
#include "path-solve.h"
#include "kmer-search.h"
#include "extension-search.h"
#include "segment-search.h"
#include "terminal.h"
#include "distant-rna.h"

#include "univdiag.h"
#include "univdiagdef.h"

#include "cmet.h"
#include "atoi.h"

#include "stage2.h"
#include "stage3.h"
#include "comp.h"


#ifdef WORDS_BIGENDIAN
#include "bigendian.h"
#endif


#ifdef HAVE_64_BIT
#ifdef LARGE_GENOMES
#else
#define DIAGONAL_ADD_QUERYPOS 1
#endif
#endif


/* Three methods for performing a multiway merge.  Need to define one below. */

#define SPEED 1


/* Note FORMULA: formulas for querypos <-> diagonal (diagterm in call to Indexdb_read) are:

plus: diagonal = position + querylength - querypos
minus: diagonal = position + querypos + index1part

For minus, the index1part is needed in call to Indexdb_read because
position is stored at beginning of plus oligomer, which corresponds to
end of minus oligomer.  As a result, we have the following formulas:

high genomic position = diagonal (corresponds to querypos =
querylength for plus, and querypos = 0 for minus)

low genomic position = diagonal - querylength (corresponds to querypos
= 0 for plus, and querypos = querylength for minus)

Holds when we use Reader_T to read from 5' end of forward query and 3'
end of revcomp query simultaneously.  If we create a queryrc sequence,
then we can use just the plus formula, and convert the query
coordinates later.

*/


#define MAX_ITER 3		/* For looping through Segment_search */

#define NO_EXTENSIONS_BEFORE_ZERO 1

#define ALLOW_MIDDLE_ALIGNMENTS 1

/* #define EXTRACT_GENOMICSEG 1 */
#ifdef EXTRACT_GENOMICSEG
#define MAX_INDEXSIZE 8
#endif


/* MAX_NALIGNMENTS of 2 vs 1 gets 1600 improvements in 275,000 reads */
/* MAX_NALIGNMENTS of 3 vs 2 gets 96 improvements in 275,000 reads */
#define MAX_NALIGNMENTS 3

#define MAX_ALLOCATION 200

#define PAIRMAX_ADDITIONAL 10000 /* Allows for finding of unpaired GMAP alignments beyond pairmax */

#define MIN_SIZELIMIT 100

static int kmer_search_sizelimit = 100;
/* static int stage1hr_sizelimit = 3000; */
/* static int extension_search_sizelimit = 3000; */

#define POLY_A 0x00000000
#define POLY_T 0xFFFFFFFF
static Oligospace_T poly_a;
static Oligospace_T poly_t;

static bool use_only_transcriptome_p;
static bool remap_transcriptome_p = false;

static Univ_IIT_T transcript_iit;
static Transcriptome_T transcriptome;
static Genome_T transcriptomebits;

static Indexdb_T indexdb_fwd;
static Indexdb_T indexdb_rev;
static Indexdb_T indexdb_tr;
static Localdb_T localdb;

static int index1part;
static int index1part_tr;
static int index1interval;

static double user_maxlevel_float;
static double user_mincoverage_float;

static Univ_IIT_T chromosome_iit;
static int circular_typeint;

static int nchromosomes;
static Genome_T genomecomp;
static Genome_T genomebits;
static Genome_T genomebits_alt;

static int leftreadshift;
static Oligospace_T oligobase_mask; /* same as kmer_mask */


/* Mode */
static Mode_T mode;
static bool snpp;
static int maxpaths_search;
static int maxpaths_report;

/* For spliceable (really "joinable", if we consider indels) */
static Chrpos_T overall_max_distance;

/* Other distances */
bool novelsplicingp;
static Chrpos_T shortsplicedist;


/* Penalties */
static int subopt_levels;

static bool find_dna_chimeras_p;
static bool distances_observed_p;

static Chrpos_T min_intronlength;
static Chrpos_T expected_pairlength;
static Chrpos_T pairlength_deviation;

static int distantsplicing_penalty;
static int min_distantsplicing_end_matches;
static int min_distantsplicing_identity;


/* GMAP parameters */
static int gmap_mode;

static bool gmap_segments_p;	/* previously called gmap_terminal_p.  Should move earlier (1). */
static bool gmap_pairsearch_p;	/* controls halfmapping.  Should move later (2). */
static bool gmap_improvement_p;	/* Should be at end (3). */

static int nullgap;
static int maxpeelback;
static int maxpeelback_distalmedial;
static int extramaterial_end;
static int extramaterial_paired;
static int trigger_score_for_gmap;
static int gmap_allowance;
static int max_gmap_pairsearch;
static int max_gmap_segments;	/* Not used */
static int max_gmap_improvement;


#define A_CHAR 0x0
#define C_CHAR 0x1
#define G_CHAR 0x2
#define T_CHAR 0x3


/* Originally allowed only 1, to print only unique translocations.
   But need to allow enough to avoid missing some translocations. */
/* For transcript splicing, need to increase MAXCHIMERAPATHS */
/* #define MAXCHIMERAPATHS 100 */
#define MAXCHIMERAPATHS 10000

#define NREQUIRED_FAST 2	/* For candidate generation using
				   multimiss.  A value of 2 implies 
				   specificity of a 24-mer, which
				   should be low for a human-sized
				   genome */

#define MAX_INDEX1INTERVAL 3
#define STAGE2_MIN_OLIGO 3	/* Actually 6, but we are adding index1interval to this */
#define GOOD_GMAP_END 6
#define GMAP_TERMINAL_TRIM 6



/* Overall flow */
#ifdef DEBUG
#define debug(x) x
#else
#define debug(x)
#endif

/* identify_segments */
#ifdef DEBUG1
#define debug1(x) x
#else
#define debug1(x)
#endif

/* Indels */ 
#ifdef DEBUG2
#define debug2(x) x
#else
#define debug2(x)
#endif

/* Indels, end */ 
#ifdef DEBUG2E
#define debug2e(x) x
#else
#define debug2e(x)
#endif

/* find_splicepairs_distant_rna */
#ifdef DEBUG4K
#define debug4k(x) x
#else
#define debug4k(x)
#endif

/* find_splicepairs_distant */
#ifdef DEBUG4L
#define debug4l(x) x
#else
#define debug4l(x)
#endif

/* find_splicepairs_distant (details) */
#ifdef DEBUG4LD
#define debug4ld(x) x
#else
#define debug4ld(x)
#endif

/* find_spliceends_shortend and find_spliceends_distant */
#ifdef DEBUG4E
#define debug4e(x) x
#else
#define debug4e(x)
#endif

/* Short overlaps */ 
#ifdef DEBUG4H
#define debug4h(x) x
#else
#define debug4h(x)
#endif

/* Pairing up segments */
#ifdef DEBUG5
#define debug5(x) x
#else
#define debug5(x)
#endif

/* Heapify */
#ifdef DEBUG6
#define debug6(x) x
#else
#define debug6(x)
#endif

/* Identify exact/onemiss/multimiss matches */
#ifdef DEBUG7
#define debug7(x) x
#else
#define debug7(x)
#endif

/* Identify onemiss matches, list contents */
#ifdef DEBUG7A
#define debug7a(x) x
#else
#define debug7a(x)
#endif

/* Filling oligos */
#ifdef DEBUG8
#define debug8(x) x
#else
#define debug8(x)
#endif

/* binary_search */
#ifdef DEBUG10
#define debug10(x) x
#else
#define debug10(x)
#endif

/* straddling at beginning of genome.  May want to turn on DEBUG11 in indexdb_hr.c */
#ifdef DEBUG11
#define debug11(x) x
#else
#define debug11(x)
#endif

/* dual_search for known splice sites */
#ifdef DEBUG12
#define debug12(x) x
#else
#define debug12(x)
#endif

/* GMAP.  May want to turn on DEBUG13 in stage3hr.c */ 
#ifdef DEBUG13
#define debug13(x) x
#else
#define debug13(x)
#endif

/* GMAP dump pairs */ 
#ifdef DEBUG13A
#define debug13a(x) x
#else
#define debug13a(x)
#endif

/* GMAP details of diagonals used */ 
#ifdef DEBUG13B
#define debug13b(x) x
#else
#define debug13b(x)
#endif

/* identify_all_segments */
#ifdef DEBUG14
#define debug14(x) x
#else
#define debug14(x)
#endif

/* binary search method for updating chrnum */
#ifdef DEBUG15
#define debug15(x) x
#else
#define debug15(x)
#endif

/* consolidate_paired_results and choose_among_paired */ 
#ifdef DEBUG16
#define debug16(x) x
#else
#define debug16(x)
#endif


#define T Stage1_T
static T
Stage1_new (int querylength) {
  T new = (T) MALLOC(sizeof(*new));
  int overhang = index1interval-1;

  new->plus_validp = (bool *) CALLOC(querylength+overhang,sizeof(bool));
  new->minus_validp = (bool *) CALLOC(querylength+overhang,sizeof(bool));

  new->plus_oligos = (Oligospace_T *) MALLOC((querylength+overhang)*sizeof(Oligospace_T));
  new->minus_oligos = (Oligospace_T *) MALLOC((querylength+overhang)*sizeof(Oligospace_T));

  new->retrievedp_allocated = (bool *) CALLOC(2 * (querylength+overhang),sizeof(bool));
  new->plus_retrievedp = &(new->retrievedp_allocated[overhang]);
  new->minus_retrievedp = &(new->retrievedp_allocated[(querylength+overhang)+overhang]);

#ifdef LARGE_GENOMES
  new->positions_high_allocated = (unsigned char **) CALLOC(2 * (querylength+overhang),sizeof(unsigned char *));
  new->plus_positions_high = &(new->positions_high_allocated[overhang]);
  new->minus_positions_high = &(new->positions_high_allocated[(querylength+overhang)+overhang]);
#endif

  new->positions_allocated = (UINT4 **) CALLOC(2 * (querylength+overhang),sizeof(UINT4 *));
  new->plus_positions = &(new->positions_allocated[overhang]);
  new->minus_positions = &(new->positions_allocated[(querylength+overhang)+overhang]);

  new->npositions_allocated = (int *) CALLOC(2 * (querylength+overhang),sizeof(int));
  new->plus_npositions = &(new->npositions_allocated[overhang]);
  new->minus_npositions = &(new->npositions_allocated[(querylength+overhang)+overhang]);

  return new;
}

#ifdef DEBUG
static void
Stage1_dump (T this, int querylength) {
  int query_lastpos = querylength - index1part, querypos;
  int i;

  for (querypos = 0; querypos <= query_lastpos; querypos++) {
    if (this->plus_retrievedp[querypos] == true) {
      printf("plus %d (%d):",querypos,this->plus_npositions[querypos]);
      for (i = 0; i < this->plus_npositions[querypos]; i++) {
	printf(" %u",this->plus_positions[querypos][i]);
      }
      printf("\n");
    }
    if (this->minus_retrievedp[querypos] == true) {
      printf("minus %d (%d):",querypos,this->minus_npositions[querypos]);
      for (i = 0; i < this->minus_npositions[querypos]; i++) {
	printf(" %u",this->minus_positions[querypos][i]);
      }
      printf("\n");
    }
  }
  printf("\n");
  return;
}
#endif


/* onep indicates whether Kmer_search_genome_one_end was run, which
   fills beyond the first index1interval */
static void
Stage1_init (T this, int querylength, bool onep,
#ifdef LARGE_GENOMES
	     unsigned char *plus_rawpositions_high_5[], unsigned char *minus_rawpositions_high_5[],
	     unsigned char *plus_rawpositions_high_3[], unsigned char *minus_rawpositions_high_3[],
#endif
	     UINT4 *plus_rawpositions_5[], UINT4 *minus_rawpositions_5[],
	     UINT4 *plus_rawpositions_3[], UINT4 *minus_rawpositions_3[],
	     int plus_nrawpositions_5[], int minus_nrawpositions_5[],
	     int plus_nrawpositions_3[], int minus_nrawpositions_3[]) {
  int query_lastpos = querylength - index1part;
  int mod;

  for (mod = 0; mod < index1interval; mod++) {
    this->plus_retrievedp[mod] = true;
#ifdef LARGE_GENOMES
    this->plus_positions_high[mod] = plus_rawpositions_high_5[mod];
#endif
    this->plus_positions[mod] = plus_rawpositions_5[mod];
    this->plus_npositions[mod] = plus_nrawpositions_5[mod];

    this->plus_retrievedp[query_lastpos-mod] = true;
#ifdef LARGE_GENOMES
    this->plus_positions_high[query_lastpos-mod] = plus_rawpositions_high_3[mod];
#endif
    this->plus_positions[query_lastpos-mod] = plus_rawpositions_3[mod];
    this->plus_npositions[query_lastpos-mod] = plus_nrawpositions_3[mod];

    /* Using new sarray and segment-based conventions */
    this->minus_retrievedp[query_lastpos-mod] = true;
#ifdef LARGE_GENOMES
    this->minus_positions_high[query_lastpos-mod] = minus_rawpositions_high_5[mod];
#endif
    this->minus_positions[query_lastpos-mod] = minus_rawpositions_5[mod];
    this->minus_npositions[query_lastpos-mod] = minus_nrawpositions_5[mod];

    this->minus_retrievedp[mod] = true;
#ifdef LARGE_GENOMES
    this->minus_positions_high[mod] = minus_rawpositions_high_3[mod];
#endif
    this->minus_positions[mod] = minus_rawpositions_3[mod];
    this->minus_npositions[mod] = minus_nrawpositions_3[mod];

#if 0
    printf("Initializing plus_positions[%d] to be %p, with count of %d\n",
	   mod,this->plus_positions[mod],this->plus_npositions[mod]);
    printf("Initializing plus_positions[%d] to be %p, with count of %d\n",
	 query_lastpos-mod,this->plus_positions[query_lastpos-mod],this->plus_npositions[query_lastpos-mod]);

    printf("Initializing minus_positions[%d] to be %p, with count of %d\n",
	   mod,this->minus_positions[mod],this->minus_npositions[mod]);
    printf("Initializing minus_positions[%d] to be %p, with count of %d\n",
	 query_lastpos-mod,this->minus_positions[query_lastpos-mod],this->minus_npositions[query_lastpos-mod]);
#endif
  }

  if (onep == true) {
    /* Not currently doing one_kmer */
    for (mod = 0; mod < index1interval; mod++) {
      this->plus_retrievedp[index1part+mod] = true;
#ifdef LARGE_GENOMES
      this->plus_positions_high[index1part+mod] = plus_rawpositions_high_5[index1interval+mod];
#endif
      this->plus_positions[index1part+mod] = plus_rawpositions_5[index1interval+mod];
      this->plus_npositions[index1part+mod] = plus_nrawpositions_5[index1interval+mod];

      this->plus_retrievedp[query_lastpos-index1part-mod] = true;
#ifdef LARGE_GENOMES
      this->plus_positions_high[query_lastpos-index1part-mod] = plus_rawpositions_high_3[index1interval+mod];
#endif
      this->plus_positions[query_lastpos-index1part-mod] = plus_rawpositions_3[index1interval+mod];
      this->plus_npositions[query_lastpos-index1part-mod] = plus_nrawpositions_3[index1interval+mod];

      /* Using new sarray and segment-based conventions */
      this->minus_retrievedp[query_lastpos-index1part-mod] = true;
#ifdef LARGE_GENOMES
      this->minus_positions_high[query_lastpos-index1part-mod] = minus_rawpositions_high_5[index1interval+mod];
#endif
      this->minus_positions[query_lastpos-index1part-mod] = minus_rawpositions_5[index1interval+mod];
      this->minus_npositions[query_lastpos-index1part-mod] = minus_nrawpositions_5[index1interval+mod];

      this->minus_retrievedp[index1part+mod] = true;
#ifdef LARGE_GENOMES
      this->minus_positions_high[index1part+mod] = minus_rawpositions_high_3[index1interval+mod];
#endif
      this->minus_positions[index1part+mod] = minus_rawpositions_3[index1interval+mod];
      this->minus_npositions[index1part+mod] = minus_nrawpositions_3[index1interval+mod];

#if 0
      printf("Initializing plus_positions[%d] to be %p, with count of %d\n",
	     index1part+mod,this->plus_positions[index1part+mod],this->plus_npositions[index1part+mod]);
      printf("Initializing plus_positions[%d] to be %p, with count of %d\n",
	     query_lastpos-index1part-mod,this->plus_positions[query_lastpos-index1part-mod],
	     this->plus_npositions[query_lastpos-index1part-mod]);

      printf("Initializing minus_positions[%d] to be %p, with count of %d\n",
	     index1part+mod,this->minus_positions[index1part+mod],this->minus_npositions[index1part+mod]);
      printf("Initializing minus_positions[%d] to be %p, with count of %d\n",
	     query_lastpos-index1part-mod,this->minus_positions[query_lastpos-index1part-mod],
	     this->minus_npositions[query_lastpos-index1part-mod]);
#endif
    }
  }

  /* Stage1_dump(this,querylength); */

  return;
}


void
Stage1_fill_all_oligos (T this, char *queryuc_ptr, int querylength, int genestrand) {
  Reader_T reader;
  int query_lastpos, querypos, querypos_rc;
  Oligostate_T last_state = INIT;
  Oligospace_T forward = 0, revcomp = 0, oligo;
#ifdef HAVE_64_BIT
  Uint8table_T plus_seenp, minus_seenp;
#else
  Uinttable_T plus_seenp, minus_seenp;
#endif

#ifdef HAVE_64_BIT
  plus_seenp = Uint8table_new(/*hint*/querylength);
  minus_seenp = Uint8table_new(/*hint*/querylength);
#else
  plus_seenp = Uinttable_new(/*hint*/querylength);
  minus_seenp = Uinttable_new(/*hint*/querylength);
#endif

  query_lastpos = querylength - index1part;
  reader = Reader_new(queryuc_ptr,/*querystart*/0,/*queryend*/querylength);

  /* Note: leftshifting is done here, rather than in Oligo_lookup */
  /* Format is 010llX because 19-mer is maximum k-mer size, which would require 10 chars */
  /* debug(printf("oligobase_mask: %010llX\n",oligobase_mask)); */
  querypos = 0;
  while ((last_state = Oligo_next_5(last_state,&querypos,&forward,&revcomp,reader,genestrand)) != DONE) {
    if (last_state != VALID) {
      /* querypos is not defined when last_state != VALID */
      debug8(printf("oligo at plus %d, minus %d is not valid\n",querypos,querypos_rc));
    } else {
      querypos_rc = query_lastpos - querypos;
      oligo = this->plus_oligos[querypos] = forward & oligobase_mask;
      debug8(printf("Putting oligo %016lX at plus %d\n",oligo,querypos));
#ifdef HAVE_64_BIT
      if (Uint8table_get(plus_seenp,oligo) != NULL) {
	debug8(printf("oligo at plus %d already seen, so marking as invalid\n",querypos));
	this->plus_validp[querypos] = false;
      } else {
	this->plus_validp[querypos] = true;
	Uint8table_put(plus_seenp,oligo,(void *) true);
      }
#else
      if (Uinttable_get(plus_seenp,oligo) != NULL) {
	debug8(printf("oligo at plus %d already seen, so marking as invalid\n",querypos));
	this->plus_validp[querypos] = false;
      } else {
	this->plus_validp[querypos] = true;
	Uinttable_put(plus_seenp,oligo,(void *) true);
      }
#endif

      oligo = this->minus_oligos[querypos_rc] = (revcomp >> leftreadshift) & oligobase_mask;
      debug8(printf("Putting oligo %016lX at minus %d\n",oligo,querypos_rc));
#ifdef HAVE_64_BIT
      if (Uint8table_get(minus_seenp,oligo) != NULL) {
	debug8(printf("oligo at minus %d already seen, so marking as invalid\n",querypos));
	this->minus_validp[querypos_rc] = false;
      } else {
	this->minus_validp[querypos_rc] = true;
	Uint8table_put(minus_seenp,oligo,(void *) true);
      }
#else
      if (Uinttable_get(minus_seenp,oligo) != NULL) {
	debug8(printf("oligo at minus %d already seen, so marking as invalid\n",querypos));
	this->minus_validp[querypos_rc] = false;
      } else {
	this->minus_validp[querypos_rc] = true;
	Uinttable_put(minus_seenp,oligo,(void *) true);
      }
#endif
    }
  }

  Reader_free(&reader);

#ifdef HAVE_64_BIT
  Uint8table_free(&minus_seenp);
  Uint8table_free(&plus_seenp);
#else
  Uinttable_free(&minus_seenp);
  Uinttable_free(&plus_seenp);
#endif

  return;
}



static void
Stage1_ptr_all_positions (T this, int querylength, int genestrand) {
  int query_lastpos, querypos, querypos_rc;
  Indexdb_T plus_indexdb, minus_indexdb;


  if (genestrand == +2) {
    plus_indexdb = indexdb_rev;
    minus_indexdb = indexdb_fwd;
  } else {
    plus_indexdb = indexdb_fwd;
    minus_indexdb = indexdb_rev;
  }

  query_lastpos = querylength - index1part;

  this->total_plus_npositions = this->total_minus_npositions = 0;

  /* Note: leftshifting is done here, rather than in Oligo_lookup */
  /* Format is 010llX because 19-mer is maximum k-mer size, which would require 10 chars */
  /* debug(printf("oligobase_mask: %010llX\n",oligobase_mask)); */
  for (querypos = 0, querypos_rc = query_lastpos; querypos <= query_lastpos; querypos++, --querypos_rc) {
    if (this->plus_retrievedp[querypos] == true) {
      /* No need to do anything */
      this->total_plus_npositions += this->plus_npositions[querypos];
    } else if (this->plus_validp[querypos] == false) {
      /* Oligo not valid */
    } else {
#ifdef LARGE_GENOMES
      this->plus_npositions[querypos] = 
	Indexdb_largeptr_with_diagterm(&this->plus_positions_high[querypos],&this->plus_positions[querypos],
				       plus_indexdb,this->plus_oligos[querypos],/*diagterm*/-querypos);
#else
      this->plus_npositions[querypos] = 
	Indexdb_ptr_with_diagterm(&this->plus_positions[querypos],plus_indexdb,
				  this->plus_oligos[querypos],/*diagterm*/-querypos);
#endif
      this->total_plus_npositions += this->plus_npositions[querypos];
      this->plus_retrievedp[querypos] = true;
    }

    if (this->minus_retrievedp[querypos_rc] == true) {
      /* No need to do anything */
      this->total_minus_npositions += this->minus_npositions[querypos_rc];
    } else if (this->minus_validp[querypos_rc] == false) {
      /* Oligo not valid */
    } else {
#ifdef LARGE_GENOMES
      this->minus_npositions[querypos_rc] =
	Indexdb_largeptr_with_diagterm(&this->minus_positions_high[querypos_rc],&this->minus_positions[querypos_rc],
				       minus_indexdb,this->minus_oligos[querypos_rc],/*diagterm*/-querypos_rc);
#else
      this->minus_npositions[querypos_rc] =
	Indexdb_ptr_with_diagterm(&this->minus_positions[querypos_rc],minus_indexdb,
				  this->minus_oligos[querypos_rc],/*diagterm*/-querypos_rc);
#endif
      this->total_minus_npositions += this->minus_npositions[querypos_rc];
      this->minus_retrievedp[querypos_rc] = true;
    }
  }

  return;
}


static int
determine_sizelimit (T this, int querylength) {
  int cutoff, *set, count;
  int n;
  int query_lastpos, querypos, querypos_rc;

  assert(querylength >= index1part);

  query_lastpos = querylength - index1part;
  set = (int *) MALLOC(2*(query_lastpos+1)*sizeof(int));
  n = 0;
  for (querypos = 0, querypos_rc = query_lastpos; querypos <= query_lastpos; querypos++, --querypos_rc) {
    if (this->plus_validp[querypos] == true) {
      set[n++] = count = this->plus_npositions[querypos];
    }

    if (this->minus_validp[querypos_rc] == true) {
      set[n++] = count = this->minus_npositions[querypos_rc];
    }
  }

  if (n < 5) {
    cutoff = MIN_SIZELIMIT;
  } else if ((cutoff = Orderstat_int_pct_inplace(set,n,/*pct*/0.60)) < MIN_SIZELIMIT) {
    cutoff = MIN_SIZELIMIT;
  }
  FREE(set);

  return cutoff;
}





void
Stage1_free (T *old) {

  /* Stage1hr_check(*old); */

  if (*old) {
#if 0
    /* Now pointing to data structure, and not copying values */
    for (i = 0; i <= query_lastpos; i++) {
      if ((*old)->plus_retrievedp[i] == true) {
#ifdef LARGE_GENOMES
	FREE((*old)->plus_positions_high[i]);
#endif
	FREE((*old)->plus_positions[i]);
      }

      if ((*old)->minus_retrievedp[i] == true) {
#ifdef LARGE_GENOMES
	FREE((*old)->minus_positions_high[i]);
#endif
	FREE((*old)->minus_positions[i]);
      }
    }
#endif

#ifdef LARGE_GENOMES
    FREE((*old)->positions_high_allocated);
#endif
    FREE((*old)->positions_allocated);
    FREE((*old)->npositions_allocated);

    FREE((*old)->retrievedp_allocated);

    FREE((*old)->minus_oligos);
    FREE((*old)->plus_oligos);

    FREE((*old)->minus_validp);
    FREE((*old)->plus_validp);

    FREE(*old);
  }

  return;
}


/************************************************************************/

static char complCode[128] = COMPLEMENT_LC;

static void
make_complement_buffered (char *complement, char *sequence, unsigned int length) {
  int i, j;

  /* complement = (char *) CALLOC(length+1,sizeof(char)); */
  for (i = length-1, j = 0; i >= 0; i--, j++) {
    complement[j] = complCode[(int) sequence[i]];
  }
  complement[length] = '\0';
  return;
}


#if 0
static void
make_complement_inplace (char *sequence, unsigned int length) {
  char temp;
  unsigned int i, j;

  for (i = 0, j = length-1; i < length/2; i++, j--) {
    temp = complCode[(int) sequence[i]];
    sequence[i] = complCode[(int) sequence[j]];
    sequence[j] = temp;
  }
  if (i == j) {
    sequence[i] = complCode[(int) sequence[i]];
  }

  return;
}
#endif

/************************************************************************/



static void
stage3list_gc (List_T *old) {
  List_T p;
  Stage3end_T hit;

  for (p = *old; p != NULL; p = p->rest) {
    hit = (Stage3end_T) p->first;
    Stage3end_free(&hit);
  }
  List_free(&(*old));
  return;
}



static List_T
single_read (char *queryuc_ptr, int querylength,
	     Compress_T query_compress_fwd, Compress_T query_compress_rev,
	     int genestrand, Oligoindex_array_T oligoindices_minor,
	     Pairpool_T pairpool, Diagpool_T diagpool, Cellpool_T cellpool,
	     Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR) {
  char *queryrc = NULL;
  int found_score;

  int sizelimit;

  int done_level, nmismatches_allowed;
  int max_terminal_mismatches, max_splice_mismatches;
  int min_trim;

  bool abort_exact_p = false;

  List_T transcripts;
  Stage3end_T hit;

  List_T p;
  List_T hits_gplus = NULL, hits_gminus = NULL;
  bool onep = false;

  char *remap_sequence;
  int remap_seqlength;

  /* Returned from Kmer_search_genome_ends and used to create stage1 for Kmer_search_genome_complete */
#ifdef LARGE_GENOMES
  unsigned char *plus_rawpositions_high_5[6], *minus_rawpositions_high_5[6],
    *plus_rawpositions_high_3[6], *minus_rawpositions_high_3[6];
#endif
  UINT4 *plus_rawpositions_5[6], *minus_rawpositions_5[6], *plus_rawpositions_3[6], *minus_rawpositions_3[6];
  int plus_nrawpositions_5[6], minus_nrawpositions_5[6], plus_nrawpositions_3[6], minus_nrawpositions_3[6];
  int plus_diagterms_5[6], minus_diagterms_5[6], plus_diagterms_3[6], minus_diagterms_3[6];

  /* Returned from Extension_search and used for distant splicing */
  List_T queryfwd_plus_set = NULL, queryfwd_minus_set = NULL, queryrev_plus_set = NULL, queryrev_minus_set = NULL;

  /* For Segment_search */
  struct Record_T *plus_all_records = NULL, *minus_all_records = NULL;
  Record_T *plus_records = NULL, *minus_records = NULL;
  int plus_nrecords = 0, minus_nrecords = 0;
  int *plus_candidates = NULL, *minus_candidates = NULL;
  int plus_ncandidates = 0, minus_ncandidates = 0;

  /* Returned from Segment_search and used in Path_solve_via_gmap for SEG_GMAP */
  List_T plus_complete_paths = NULL, minus_complete_paths = NULL;

  Stage1_T stage1 = NULL;
  int mod;


  debug(printf("Entered single_read with queryuc_ptr %s\n",queryuc_ptr));

  if (user_maxlevel_float < 0.0) {
    nmismatches_allowed = querylength/index1part;
  } else if (user_maxlevel_float > 0.0 && user_maxlevel_float < 1.0) {
    nmismatches_allowed = (int) rint(user_maxlevel_float * (double) querylength);
  } else {
    nmismatches_allowed = (int) user_maxlevel_float;
  }

  found_score = querylength;

  if (indexdb_tr != NULL) {
    debug(printf("Running Kmer_search_transcriptome_single\n"));
    Kmer_search_transcriptome_single(&found_score,&hits_gplus,&hits_gminus,
				     queryuc_ptr,querylength,
				     query_compress_fwd,query_compress_rev,
				     transcript_iit,transcriptome,transcriptomebits,
				     nmismatches_allowed,kmer_search_sizelimit,/*level*/0);

  }

  if (use_only_transcriptome_p == false) {

    if (found_score > nmismatches_allowed) {
      debug(printf("LEVEL %d: RUNNING KMER_SEARCH_GENOME_ENDS\n",1));
      for (mod = 0; mod < 2*index1interval; mod++) {
#ifdef LARGE_GENOMES
	plus_rawpositions_high_5[mod] = (unsigned char *) NULL;
	minus_rawpositions_high_5[mod] = (unsigned char *) NULL;
	plus_rawpositions_high_3[mod] = (unsigned char *) NULL;
	minus_rawpositions_high_3[mod] = (unsigned char *) NULL;
#endif
	plus_rawpositions_5[mod] = (UINT4 *) NULL;
	minus_rawpositions_5[mod] = (UINT4 *) NULL;
	plus_rawpositions_3[mod] = (UINT4 *) NULL;
	minus_rawpositions_3[mod] = (UINT4 *) NULL;
      }
      Kmer_search_genome_ends_exact(&abort_exact_p,&found_score,&hits_gplus,&hits_gminus,

#ifdef LARGE_GENOMES
				    plus_rawpositions_high_5,minus_rawpositions_high_5,
				    plus_rawpositions_high_3,minus_rawpositions_high_3,
#endif
				    plus_rawpositions_5,plus_nrawpositions_5,plus_diagterms_5,
				    minus_rawpositions_5,minus_nrawpositions_5,minus_diagterms_5,
				    plus_rawpositions_3,plus_nrawpositions_3,plus_diagterms_3,
				    minus_rawpositions_3,minus_nrawpositions_3,minus_diagterms_3,

				    queryuc_ptr,querylength,
				    query_compress_fwd,query_compress_rev,
				    genestrand,nmismatches_allowed,/*max_hits*/1000,
				    /*level*/1);
      
      debug(printf("After Kmer_search_genome_ends, we have found_score %d (vs allowed %d), %d plus and %d minus hits\n",
		   found_score,nmismatches_allowed,List_length(hits_gplus),List_length(hits_gminus)));
    }
    
    if (found_score > nmismatches_allowed && querylength >= index1part) {
      stage1 = Stage1_new(querylength);
      Stage1_init(stage1,querylength,onep,
#ifdef LARGE_GENOMES
		  plus_rawpositions_high_5,minus_rawpositions_high_5,
		  plus_rawpositions_high_3,minus_rawpositions_high_3,
#endif
		  plus_rawpositions_5,minus_rawpositions_5,
		  plus_rawpositions_3,minus_rawpositions_3,
		  plus_nrawpositions_5,minus_nrawpositions_5,
		  plus_nrawpositions_3,minus_nrawpositions_3);
      Stage1_fill_all_oligos(stage1,queryuc_ptr,querylength,genestrand);

      queryrc = (char *) MALLOC((querylength+1)*sizeof(char));
      make_complement_buffered(queryrc,queryuc_ptr,querylength);

      debug(printf("LEVEL %d: RUNNING EXTENSION_SEARCH\n",2));
      Extension_search(&found_score,&hits_gplus,&hits_gminus,
		       &queryfwd_plus_set,&queryfwd_minus_set,&queryrev_plus_set,&queryrev_minus_set,
		       stage1,queryuc_ptr,queryrc,querylength,query_compress_fwd,query_compress_rev,
		       pairpool,dynprogL,dynprogM,dynprogR,oligoindices_minor,diagpool,cellpool,
		       nmismatches_allowed,genestrand,/*paired_end_p*/false,/*level*/2);
      
    }

    if (found_score > nmismatches_allowed && abort_exact_p == false) {
      if (queryrc == NULL) {
	queryrc = (char *) MALLOC((querylength+1)*sizeof(char));
	make_complement_buffered(queryrc,queryuc_ptr,querylength);
      }

      debug(printf("LEVEL %d: RUNNING KMER_SEARCH_GENOME_ENDS_APPROX\n",3));
      Kmer_search_genome_ends_approx(&found_score,&hits_gplus,&hits_gminus,
#ifdef LARGE_GENOMES
				     plus_rawpositions_high_5,minus_rawpositions_high_5,
				     plus_rawpositions_high_3,minus_rawpositions_high_3,
#endif
				     plus_rawpositions_5,plus_nrawpositions_5,plus_diagterms_5,
				     minus_rawpositions_5,minus_nrawpositions_5,minus_diagterms_5,
				     plus_rawpositions_3,plus_nrawpositions_3,plus_diagterms_3,
				     minus_rawpositions_3,minus_nrawpositions_3,minus_diagterms_3,

				     query_compress_fwd,query_compress_rev,querylength,
				     genestrand,nmismatches_allowed,/*max_hits*/1000,
				     /*sizelimit*/3000,/*level*/3);
    }

    if (found_score > nmismatches_allowed) {
      if (querylength < index1part) {
	/* Skip */
      } else {
	if (stage1 == NULL) {
	  stage1 = Stage1_new(querylength);
	  Stage1_init(stage1,querylength,onep,
#ifdef LARGE_GENOMES
		      plus_rawpositions_high_5,minus_rawpositions_high_5,
		      plus_rawpositions_high_3,minus_rawpositions_high_3,
#endif
		      plus_rawpositions_5,minus_rawpositions_5,
		      plus_rawpositions_3,minus_rawpositions_3,
		      plus_nrawpositions_5,minus_nrawpositions_5,
		      plus_nrawpositions_3,minus_nrawpositions_3);
	  Stage1_fill_all_oligos(stage1,queryuc_ptr,querylength,genestrand);
	}
	Stage1_ptr_all_positions(stage1,querylength,genestrand);
	/* Need sizelimit to constrain segment search */
	sizelimit = determine_sizelimit(stage1,querylength);

	if (queryrc == NULL) {
	  queryrc = (char *) MALLOC((querylength+1)*sizeof(char));
	  make_complement_buffered(queryrc,queryuc_ptr,querylength);
	}

	plus_records = Segment_identify(&plus_nrecords,&plus_all_records,
					&plus_candidates,&plus_ncandidates,
#ifdef LARGE_GENOMES
					stage1->plus_positions_high,
#endif
					stage1->plus_positions,stage1->plus_npositions,stage1->plus_validp,
					querylength,sizelimit,/*plusp*/true);
	minus_records = Segment_identify(&minus_nrecords,&minus_all_records,
					 &minus_candidates,&minus_ncandidates,
#ifdef LARGE_GENOMES
					 stage1->minus_positions_high,
#endif
					 stage1->minus_positions,stage1->minus_npositions,stage1->minus_validp,
					 querylength,sizelimit,/*plusp*/false);

	Segment_search_all(&found_score,&hits_gplus,&hits_gminus,
			   &plus_complete_paths,&minus_complete_paths,
			   
			   plus_records,plus_nrecords,plus_candidates,plus_ncandidates,
			   minus_records,minus_nrecords,minus_candidates,minus_ncandidates,
			   
			   queryuc_ptr,queryrc,querylength,
			   query_compress_fwd,query_compress_rev,
			   genestrand,/*paired_end_p*/false,/*level*/4);

	FREE(minus_candidates);
	FREE(plus_candidates);
	
	FREE(minus_records);
	FREE(plus_records);
	
	FREE(minus_all_records);
	FREE(plus_all_records);
      }

      Univdiag_list_gc(&plus_complete_paths);
      Univdiag_list_gc(&minus_complete_paths);
    }

    Stage1_free(&stage1);


    /* Remap to transcriptome */
    if (indexdb_tr != NULL && remap_transcriptome_p == true) {
      for (p = hits_gplus; p != NULL; p = List_next(p)) {
	hit = (Stage3end_T) List_head(p);
	if (Stage3end_transcripts(hit) != NULL) {
	  /* Already mapped to transcriptome */
	} else {
	  remap_sequence = Stage3end_substrings_genomic_sequence(&remap_seqlength,hit,genomecomp);
	  if ((transcripts = Kmer_remap_transcriptome(remap_sequence,remap_seqlength,
						      Stage3end_chrnum(hit),Stage3end_chrpos_low(hit),Stage3end_chrpos_high(hit),
						      transcript_iit,transcriptomebits,
						      transcriptome,genomecomp,kmer_search_sizelimit)) != NULL) {
	    /* Following procedure frees the list hit->transcripts */
	    Stage3end_set_transcripts(hit,transcripts);
	  }
	  FREE(remap_sequence);
	}
      }

      for (p = hits_gminus; p != NULL; p = List_next(p)) {
	hit = (Stage3end_T) List_head(p);
	if (Stage3end_transcripts(hit) != NULL) {
	  /* Already mapped to transcriptome */
	} else {
	  remap_sequence = Stage3end_substrings_genomic_sequence(&remap_seqlength,hit,genomecomp);
	  if ((transcripts = Kmer_remap_transcriptome(remap_sequence,remap_seqlength,
						      Stage3end_chrnum(hit),Stage3end_chrpos_low(hit),Stage3end_chrpos_high(hit),
						      transcript_iit,transcriptomebits,
						      transcriptome,genomecomp,kmer_search_sizelimit)) != NULL) {
	    /* Following procedure frees the list hit->transcripts */
	    Stage3end_set_transcripts(hit,transcripts);
	  }
	  FREE(remap_sequence);
	}
      }
    }
    


    /* Distant splicing */
    if (found_score > nmismatches_allowed) {
      if ((done_level = found_score + subopt_levels) > nmismatches_allowed) {
	done_level = nmismatches_allowed;
      }
      debug(printf("done_level %d = found_score %d + subopt_levels %d\n",done_level,found_score,subopt_levels));

      min_trim = querylength;	/* min of max(trim5,trim3) over each hit */
      for (p = hits_gplus; p != NULL; p = List_next(p)) {
	hit = (Stage3end_T) List_head(p);
	if (Stage3end_max_trim(hit) < min_trim) {
	  min_trim = Stage3end_max_trim(hit);
	}
      }
      for (p = hits_gminus; p != NULL; p = List_next(p)) {
	hit = (Stage3end_T) List_head(p);
	if (Stage3end_max_trim(hit) < min_trim) {
	  min_trim = Stage3end_max_trim(hit);
	}
      }
      
      debug(printf("Comparing min_trim %d with min_distantsplicing_end_matches %d\n",
		   min_trim,min_distantsplicing_end_matches));
      if (min_trim < min_distantsplicing_end_matches) {
	/* Don't find distant splicing */
	debug(printf("Skipping distant splicing because min_trim %d < min_distantsplicing_end_matches %d\n",
		     min_trim,min_distantsplicing_end_matches));
      
      } else if ((max_splice_mismatches = done_level - distantsplicing_penalty) >= 0) {
	if (novelsplicingp == true) {
	  debug(printf("Candidate for distant splicing because min_trim %d >= min_distantsplicing_end_matches %d\n",
		       min_trim,min_distantsplicing_end_matches));
	  debug(printf("Have sets: %d %d %d %d\n",
		       List_length(queryfwd_plus_set),List_length(queryfwd_minus_set),
		       List_length(queryrev_plus_set),List_length(queryrev_minus_set)));
	  Distant_rna_solve(&found_score,&hits_gplus,&hits_gminus,
					 
			    queryfwd_plus_set,queryfwd_minus_set,
			    queryrev_plus_set,queryrev_minus_set,
					 
			    query_compress_fwd,query_compress_rev,querylength,
			    /*user_maxlevel*/done_level,max_splice_mismatches,
			    genestrand,/*first_read_p*/true,/*level*/6);
	} else {
	  /* TODO: Implement distant DNA fusions */
	}
      }
    }
  

    /* Terminals.  Use hits == NULL as criterion, rather than
       found_score, since Terminal_solve cannot improve found_score */
    if (hits_gplus == NULL && hits_gminus == NULL) {
      /* No need to recompute min_trim, if hits_gplus and hits_gminus are NULL */

      max_terminal_mismatches = done_level;
      if ((hits_gplus = Terminal_solve_plus(&found_score,
					    queryfwd_plus_set,queryrev_plus_set,
					    query_compress_fwd,querylength,
					    /*user_maxlevel*/done_level,max_terminal_mismatches,
					    genestrand,/*first_read_p*/true,/*level*/7)) == NULL) {
	debug(printf("Got no terminals\n"));
      } else {
	debug(printf("Got %d terminals\n",List_length(hits_gplus)));
      }

      if ((hits_gminus = Terminal_solve_minus(&found_score,
					      queryfwd_minus_set,queryrev_minus_set,
					      query_compress_rev,querylength,
					      /*user_maxlevel*/done_level,max_terminal_mismatches,
					      genestrand,/*first_read_p*/true,/*level*/7)) == NULL) {
	debug(printf("Got no terminals\n"));
      } else {
	debug(printf("Got %d terminals\n",List_length(hits_gminus)));
      }
    }

    Elt_gc(&queryfwd_plus_set);
    Elt_gc(&queryfwd_minus_set);
    Elt_gc(&queryrev_plus_set);
    Elt_gc(&queryrev_minus_set);

    FREE(queryrc);
  }

  return List_append(hits_gplus,hits_gminus);
}


Stage3end_T *
Stage1_single_read (int *npaths_primary, int *npaths_altloc, int *first_absmq, int *second_absmq,
		    Shortread_T queryseq, Oligoindex_array_T oligoindices_minor,
		    Pairpool_T pairpool, Diagpool_T diagpool, Cellpool_T cellpool,
		    Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR) {
  Stage3end_T *stage3array;
  List_T hits, geneplus_hits, geneminus_hits;

  int querylength, min_coverage;
  char *queryuc_ptr, *quality_string;
  Compress_T query_compress_fwd, query_compress_rev;


  querylength = Shortread_fulllength(queryseq);
  queryuc_ptr = Shortread_fullpointer_uc(queryseq);
  query_compress_fwd = Compress_new_fwd(queryuc_ptr,querylength);
  query_compress_rev = Compress_new_rev(queryuc_ptr,querylength);

  if (mode == STANDARD || mode == CMET_STRANDED || mode == ATOI_STRANDED || mode == TTOC_STRANDED) {
    hits = single_read(queryuc_ptr,querylength,query_compress_fwd,query_compress_rev,
		       /*genestrand*/0,oligoindices_minor,
		       pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);


  } else if (mode == CMET_NONSTRANDED || mode == ATOI_NONSTRANDED || mode == TTOC_NONSTRANDED) {
    geneplus_hits = single_read(queryuc_ptr,querylength,
				query_compress_fwd,query_compress_rev,
				/*genestrand*/+1,oligoindices_minor,
				pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

    geneminus_hits = single_read(queryuc_ptr,querylength,
				 query_compress_fwd,query_compress_rev,
				 /*genestrand*/+2,oligoindices_minor,
				 pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

    hits = List_append(geneplus_hits,geneminus_hits);

  } else {
    fprintf(stderr,"Do not recognize mode %d\n",mode);
    abort();
  }

  if (user_mincoverage_float < 0.0) {
    min_coverage = 0;
  } else if (user_mincoverage_float > 0.0 && user_mincoverage_float < 1.0) {
    min_coverage = (int) rint(user_mincoverage_float * (double) querylength);
  } else {
    min_coverage = (int) user_mincoverage_float;
  }

  debug(printf("Have %d hits before filter_coverage\n",List_length(hits)));
  hits = Stage3end_filter_coverage(hits,min_coverage);
  debug(printf("Have %d hits after filter_coverage\n",List_length(hits)));

  if (hits == NULL) {
    *npaths_primary = *npaths_altloc = 0;
    stage3array = (Stage3end_T *) NULL;
    
  } else {
    hits = Stage3end_remove_circular_alias(hits); /* Contains a call to unalias_circular */
    hits = Stage3end_remove_duplicates(hits); /* Aliases can cause duplicates */
    
    hits = Stage3end_optimal_score(hits,querylength,/*finalp*/false);
    hits = Stage3end_remove_overlaps(hits,/*finalp*/true);
    hits = Stage3end_optimal_score(hits,querylength,/*finalp*/true);

    quality_string = Shortread_quality_string(queryseq);
    Stage3end_count_hits(&(*npaths_primary),&(*npaths_altloc),hits);
    stage3array = (Stage3end_T *) List_to_array_out(hits,NULL); List_free(&hits); /* Return value */
    stage3array = Stage3end_eval_and_sort(/*npaths*/(*npaths_primary) + (*npaths_altloc),
					  &(*first_absmq),&(*second_absmq),
					  stage3array,queryuc_ptr,quality_string,/*displayp*/true);
  }
  
  Compress_free(&query_compress_fwd);
  Compress_free(&query_compress_rev);

  return stage3array;
}




static Pairtype_T
choose_among_paired (int *best_nmatches_paired, int *best_nmatches_5, int *best_nmatches_3,
		     List_T hitpairs, List_T samechr, List_T conc_transloc) {
  Pairtype_T final_pairtype = UNPAIRED;
  List_T p;
  Stage3pair_T hitpair;
  int nmatches, nmatches5, nmatches3;

  debug16(printf("choose: %d hitpairs, %d conc_transloc, %d samechr\n",
		 List_length(hitpairs),List_length(conc_transloc),List_length(samechr)));

  *best_nmatches_paired = 0;
  *best_nmatches_5 = 0;
  *best_nmatches_3 = 0;
  for (p = hitpairs; p != NULL; p = p->rest) {
    hitpair = (Stage3pair_T) p->first;
    if ((nmatches = Stage3pair_nmatches_posttrim(&nmatches5,&nmatches3,hitpair)) > *best_nmatches_paired) {
      final_pairtype = CONCORDANT;
      *best_nmatches_paired = nmatches;
      *best_nmatches_5 = nmatches5;
      *best_nmatches_3 = nmatches3;
    }
  }

  *best_nmatches_paired += 1; /* penalty for choosing translocation over others */

  for (p = conc_transloc; p != NULL; p = p->rest) {
    hitpair = (Stage3pair_T) p->first;
    if ((nmatches = Stage3pair_nmatches_posttrim(&nmatches5,&nmatches3,hitpair)) > *best_nmatches_paired) {
      final_pairtype = CONCORDANT_TRANSLOCATIONS;
      *best_nmatches_paired = nmatches;
      *best_nmatches_5 = nmatches5;
      *best_nmatches_3 = nmatches3;
    }
  }

  for (p = samechr; p != NULL; p = p->rest) {
    hitpair = (Stage3pair_T) p->first;
    if ((nmatches = Stage3pair_nmatches_posttrim(&nmatches5,&nmatches3,hitpair)) > *best_nmatches_paired) {
      final_pairtype = PAIRED_UNSPECIFIED;
      *best_nmatches_paired = nmatches;
      *best_nmatches_5 = nmatches5;
      *best_nmatches_3 = nmatches3;
    }
  }

  debug16(printf("best_nmatches_paired among paired = %d = %d + %d\n",
		 *best_nmatches_paired,*best_nmatches_5,*best_nmatches_3));
  debug16(printf("final pairtype: %s\n",Pairtype_string(final_pairtype)));

  return final_pairtype;
}


#if 0
static int
best_nmatches_singleend (List_T hits) {
  int best_nmatches = 0;
  List_T p;
  Stage3end_T hit;
  int nmatches;

  for (p = hits; p != NULL; p = p->rest) {
    hit = (Stage3end_T) p->first;
    if ((nmatches = Stage3end_nmatches_posttrim(hit)) > best_nmatches) {
      best_nmatches = nmatches;
    }
  }

  return best_nmatches;
}
#endif


/* Clean up all previous calculations */
static void
paired_results_free (List_T hitpairs, List_T samechr, List_T conc_transloc,
		     List_T hits5, List_T hits3, int querylength5, int querylength3) {
  List_T p;
  Stage3pair_T stage3pair;

  for (p = hitpairs; p != NULL; p = List_next(p)) {
    stage3pair = (Stage3pair_T) List_head(p);
    Stage3pair_free(&stage3pair);
  }
  List_free(&hitpairs);

  for (p = samechr; p != NULL; p = List_next(p)) {
    stage3pair = (Stage3pair_T) List_head(p);
    Stage3pair_free(&stage3pair);
  }
  List_free(&samechr);

  for (p = conc_transloc; p != NULL; p = List_next(p)) {
    stage3pair = (Stage3pair_T) List_head(p);
    Stage3pair_free(&stage3pair);
  }
  List_free(&conc_transloc);

  stage3list_gc(&hits3);
  stage3list_gc(&hits5);

  return;
}


static List_T
paired_read (bool *abort_pairing_p, List_T *hits5, List_T *hits3, List_T *samechr, List_T *conc_transloc,
	     char *queryuc_ptr_5, int querylength5, char *queryuc_ptr_3, int querylength3,
	     Compress_T query5_compress_fwd, Compress_T query5_compress_rev,
	     Compress_T query3_compress_fwd, Compress_T query3_compress_rev,
	     int genestrand, Oligoindex_array_T oligoindices_minor,
	     Pairpool_T pairpool, Diagpool_T diagpool, Cellpool_T cellpool,
	     Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR) {
  List_T hitpairs;
  char *queryrc5 = NULL, *queryrc3 = NULL;

  /* int sensedir; -- Used by Pair_solve_via_gmap */

  /* int plus_specifici_5[3], minus_specifici_5[3], plus_specifici_3[3], minus_specifici_3[3]; */
  int sizelimit5, sizelimit3;

  /* concordant_score is for all hitpairs; adjacent score is for those hitpairs within adjacent_pairlength */
  int adjacent_score, concordant_score;
  int found_score_5, found_score_3, orig_found_score_5, orig_found_score_3, 
    done_level_5, done_level_3, nmismatches_allowed, nmismatches_allowed_5, nmismatches_allowed_3;
  int max_terminal_mismatches, max_splice_mismatches;
  int min_trim_5, min_trim_3;

  bool abort_exact5_p = false, abort_exact3_p = false;
  int maxpairedpaths;

  List_T hits5_gplus = NULL, hits5_gminus = NULL, hits3_gplus = NULL, hits3_gminus = NULL;
  Ladder_T ladder5_plus, ladder5_minus, ladder3_plus, ladder3_minus;

#if 0
  /* If we prioritize transcriptome hits over genomic hits */
  List_T transcriptome_hits5 = NULL, transcriptome_hits3 = NULL;
  List_T genome_hits5 = NULL, genome_hits5 = NULL;
#endif

  /* Returned from Kmer_search_genome_ends and used to create stage1 for Kmer_search_genome_complete */
#ifdef LARGE_GENOMES
  unsigned char *plus5_rawpositions_high_5[6], *minus5_rawpositions_high_5[6],
    *plus5_rawpositions_high_3[6], *minus5_rawpositions_high_3[6],
    *plus3_rawpositions_high_5[6], *minus3_rawpositions_high_5[6],
    *plus3_rawpositions_high_3[6], *minus3_rawpositions_high_3[6];
#endif
  UINT4 *plus5_rawpositions_5[6], *minus5_rawpositions_5[6], *plus5_rawpositions_3[6], *minus5_rawpositions_3[6];
  int plus5_nrawpositions_5[6], minus5_nrawpositions_5[6], plus5_nrawpositions_3[6], minus5_nrawpositions_3[6];
  int plus5_diagterms_5[6], minus5_diagterms_5[6], plus5_diagterms_3[6], minus5_diagterms_3[6];

  UINT4 *plus3_rawpositions_5[6], *minus3_rawpositions_5[6], *plus3_rawpositions_3[6], *minus3_rawpositions_3[6];
  int plus3_nrawpositions_5[6], minus3_nrawpositions_5[6], plus3_nrawpositions_3[6], minus3_nrawpositions_3[6];
  int plus3_diagterms_5[6], minus3_diagterms_5[6], plus3_diagterms_3[6], minus3_diagterms_3[6];

  /* Returned from Extension_search and used for distant splicing */
  List_T queryfwd_plus_set_5 = NULL, queryfwd_minus_set_5 = NULL, queryrev_plus_set_5 = NULL, queryrev_minus_set_5 = NULL,
    queryfwd_plus_set_3 = NULL, queryfwd_minus_set_3 = NULL, queryrev_plus_set_3 = NULL, queryrev_minus_set_3 = NULL;

  /* For Segment_search */
  int niter, score_start_5, score_start_3;
  struct Record_T *plus_all_records5 = NULL, *minus_all_records5 = NULL, *plus_all_records3 = NULL, *minus_all_records3 = NULL;
  Record_T *plus_records5 = NULL, *minus_records5 = NULL, *plus_records3 = NULL, *minus_records3 = NULL;
  int plus_nrecords5 = 0, minus_nrecords5 = 0, plus_nrecords3 = 0, minus_nrecords3 = 0;
  int *plus_candidates5 = NULL, *minus_candidates5 = NULL, *plus_candidates3 = NULL, *minus_candidates3 = NULL;
  int plus_ncandidates5 = 0, minus_ncandidates5 = 0, plus_ncandidates3 = 0, minus_ncandidates3 = 0;
  Intlist_T plus_anchors5 = NULL, minus_anchors5 = NULL, plus_anchors3 = NULL, minus_anchors3 = NULL;

  /* Returned from Segment_search and used in Path_solve_via_gmap for SEG_GMAP */
  List_T plus_complete_paths_5 = NULL, minus_complete_paths_5 = NULL, plus_complete_paths_3 = NULL, minus_complete_paths_3 = NULL;

  Stage1_T this5 = NULL, this3 = NULL;
  int mod;
  int level = 0;


  debug(printf("Entered paired_read with queryuc_ptr_5 %s\n",queryuc_ptr_5));
  debug(printf("Entered paired_read with queryuc_ptr_3 %s\n",queryuc_ptr_3));

  *abort_pairing_p = false;
  hitpairs = (List_T) NULL;
  *samechr = (List_T) NULL;
  *conc_transloc = (List_T) NULL;


  /* Take the larger of maxpaths_search and 10*maxpaths_report */
  maxpairedpaths = maxpaths_search;
  if (maxpairedpaths < 10*maxpaths_report) {
    maxpairedpaths = 10*maxpaths_report;
  }

  adjacent_score = concordant_score = querylength5 + querylength3;

  if (user_maxlevel_float < 0.0) {
    nmismatches_allowed_5 = querylength5/index1part_tr;
    nmismatches_allowed_3 = querylength3/index1part_tr;
  } else if (user_maxlevel_float > 0.0 && user_maxlevel_float < 1.0) {
    nmismatches_allowed_5 = (int) rint(user_maxlevel_float * (double) querylength5);
    nmismatches_allowed_3 = (int) rint(user_maxlevel_float * (double) querylength3);
  } else {
    nmismatches_allowed_5 = nmismatches_allowed_3 = (int) user_maxlevel_float;
  }
  nmismatches_allowed = nmismatches_allowed_5 + nmismatches_allowed_3;


  found_score_5 = querylength5;
  found_score_3 = querylength3;
  /* found_score = querylength5 + querylength3; */

  if (indexdb_tr == NULL) {
    ladder5_plus = Ladder_new(NULL,/*end5p*/true); hits5_gplus = (List_T) NULL;
    ladder5_minus = Ladder_new(NULL,/*end5p*/true); hits5_gminus = (List_T) NULL;
    ladder3_plus = Ladder_new(NULL,/*end5p*/false); hits3_gplus = (List_T) NULL;
    ladder3_minus = Ladder_new(NULL,/*end5p*/false); hits3_gminus = (List_T) NULL;

  } else {
    debug(printf("Running Kmer_search_transcriptome_paired\n"));
    Kmer_search_transcriptome_paired(&found_score_5,&found_score_3,
				     &hits5_gplus,&hits5_gminus,&hits3_gplus,&hits3_gminus,
				     queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				     query5_compress_fwd,query5_compress_rev,
				     query3_compress_fwd,query3_compress_rev,

				     transcript_iit,transcriptome,transcriptomebits, 
				     nmismatches_allowed_5,nmismatches_allowed_3,
				     kmer_search_sizelimit,level);

    debug(printf("Pairing up concordant hits\n"));

    ladder5_plus = Ladder_new(hits5_gplus,/*end5p*/true); hits5_gplus = (List_T) NULL;
    ladder5_minus = Ladder_new(hits5_gminus,/*end5p*/true); hits5_gminus = (List_T) NULL;
    ladder3_plus = Ladder_new(hits3_gplus,/*end5p*/false); hits3_gplus = (List_T) NULL;
    ladder3_minus = Ladder_new(hits3_gminus,/*end5p*/false); hits3_gminus = (List_T) NULL;

    hitpairs = Concordance_pair_up_transcriptome(&(*abort_pairing_p),&concordant_score,hitpairs,
						 ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
						 maxpairedpaths,genestrand);
    adjacent_score = concordant_score; /* Transcripts give a better definition than adjacency in the genome */

    level += 1;
    debug(printf("Level %d> hitpairs has %d pairs\n",level,List_length(hitpairs)));
  }

  if (use_only_transcriptome_p == false) {

    if (*abort_pairing_p == false && adjacent_score > nmismatches_allowed) {
      debug(printf("LEVEL %d: RUNNING KMER_SEARCH_GENOME_ENDS ON 5' READ\n",level));
      for (mod = 0; mod < 2*index1interval; mod++) {
#ifdef LARGE_GENOMES
	plus5_rawpositions_high_5[mod] = (unsigned char *) NULL;
	minus5_rawpositions_high_5[mod] = (unsigned char *) NULL;
	plus5_rawpositions_high_3[mod] = (unsigned char *) NULL;
	minus5_rawpositions_high_3[mod] = (unsigned char *) NULL;
#endif
	plus5_rawpositions_5[mod] = (UINT4 *) NULL;
	minus5_rawpositions_5[mod] = (UINT4 *) NULL;
	plus5_rawpositions_3[mod] = (UINT4 *) NULL;
	minus5_rawpositions_3[mod] = (UINT4 *) NULL;
      }

      Kmer_search_genome_ends_exact(&abort_exact5_p,&found_score_5,&hits5_gplus,&hits5_gminus,
#ifdef LARGE_GENOMES
				    plus5_rawpositions_high_5,minus5_rawpositions_high_5,
				    plus5_rawpositions_high_3,minus5_rawpositions_high_3,
#endif
				    plus5_rawpositions_5,plus5_nrawpositions_5,plus5_diagterms_5,
				    minus5_rawpositions_5,minus5_nrawpositions_5,minus5_diagterms_5,
				    plus5_rawpositions_3,plus5_nrawpositions_3,plus5_diagterms_3,
				    minus5_rawpositions_3,minus5_nrawpositions_3,minus5_diagterms_3,

				    queryuc_ptr_5,querylength5,
				    query5_compress_fwd,query5_compress_rev,
				    genestrand,nmismatches_allowed_5,/*max_hits*/1000,level);


      debug(printf("LEVEL %d: RUNNING KMER_SEARCH_GENOME_ENDS ON 3' READ\n",level));
      for (mod = 0; mod < 2*index1interval; mod++) {
#ifdef LARGE_GENOMES
	plus3_rawpositions_high_5[mod] = (unsigned char *) NULL;
	minus3_rawpositions_high_5[mod] = (unsigned char *) NULL;
	plus3_rawpositions_high_3[mod] = (unsigned char *) NULL;
	minus3_rawpositions_high_3[mod] = (unsigned char *) NULL;
#endif
	plus3_rawpositions_5[mod] = (UINT4 *) NULL;
	minus3_rawpositions_5[mod] = (UINT4 *) NULL;
	plus3_rawpositions_3[mod] = (UINT4 *) NULL;
	minus3_rawpositions_3[mod] = (UINT4 *) NULL;
      }

      Kmer_search_genome_ends_exact(&abort_exact3_p,&found_score_3,&hits3_gplus,&hits3_gminus,
#ifdef LARGE_GENOMES
				    plus3_rawpositions_high_5,minus3_rawpositions_high_5,
				    plus3_rawpositions_high_3,minus3_rawpositions_high_3,
#endif
				    plus3_rawpositions_5,plus3_nrawpositions_5,plus3_diagterms_5,
				    minus3_rawpositions_5,minus3_nrawpositions_5,minus3_diagterms_5,
				    plus3_rawpositions_3,plus3_nrawpositions_3,plus3_diagterms_3,
				    minus3_rawpositions_3,minus3_nrawpositions_3,minus3_diagterms_3,

				    queryuc_ptr_3,querylength3,
				    query3_compress_fwd,query3_compress_rev,
				    genestrand,nmismatches_allowed_3,/*max_hits*/1000,level);

      /* Check for concordance */
      debug(printf("found scores: %d and %d (vs %d and %d allowed)\n",
		   found_score_5,found_score_3,nmismatches_allowed_5,nmismatches_allowed_3));
      debug(printf("hits5 plus: %d new.  hits5 minus: %d new.\n",List_length(hits5_gplus),List_length(hits5_gminus)));
      debug(printf("hits3 plus: %d new.  hits3 minus: %d new.\n",List_length(hits3_gplus),List_length(hits3_gminus)));
      hitpairs = Concordance_pair_up_genome(&(*abort_pairing_p),&adjacent_score,&concordant_score,hitpairs,
					    hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					    ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					    querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(1) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		   level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
    }

    /* BOTH_ENDS exact solves 47% as concordant uniq and 35% as halfmapping uniq at 8000 queries/sec */

    /* Next step: ONE_END solves 56% of remainder as concordant uniq at 2000 queries/sec */
    /* Next step: APPROX (with merging) solves 29% of remainder as concordant uniq at 4000 queries/sec */
    /* Next step: EXT solves 94% of remainder as concordant uniq at 1900 queries/sec (wins) */

    if (*abort_pairing_p == false && adjacent_score > nmismatches_allowed) {
      if (found_score_5 >= found_score_3 && querylength5 >= index1part) {
	this5 = Stage1_new(querylength5);
	Stage1_init(this5,querylength5,/*one5p*/false,
#ifdef LARGE_GENOMES
		    plus5_rawpositions_high_5,minus5_rawpositions_high_5,
		    plus5_rawpositions_high_3,minus5_rawpositions_high_3,
#endif
		    plus5_rawpositions_5,minus5_rawpositions_5,
		    plus5_rawpositions_3,minus5_rawpositions_3,
		    plus5_nrawpositions_5,minus5_nrawpositions_5,
		    plus5_nrawpositions_3,minus5_nrawpositions_3);
	Stage1_fill_all_oligos(this5,queryuc_ptr_5,querylength5,genestrand);

	queryrc5 = (char *) MALLOC((querylength5+1)*sizeof(char));
	make_complement_buffered(queryrc5,queryuc_ptr_5,querylength5);

	debug(printf("LEVEL %d: RUNNING EXTENSION_SEARCH ON 5' READ\n",level));
	Extension_search(&found_score_5,&hits5_gplus,&hits5_gminus,
			 &queryfwd_plus_set_5,&queryfwd_minus_set_5,&queryrev_plus_set_5,&queryrev_minus_set_5,
			 this5,queryuc_ptr_5,queryrc5,querylength5,query5_compress_fwd,query5_compress_rev,
			 pairpool,dynprogL,dynprogM,dynprogR,oligoindices_minor,diagpool,cellpool,
			 nmismatches_allowed_5,genestrand,/*paired_end_p*/true,level);
      }

      if (found_score_3 >= found_score_5 && querylength3 >= index1part) {
	this3 = Stage1_new(querylength3);
	Stage1_init(this3,querylength3,/*one3p*/false,
#ifdef LARGE_GENOMES
		    plus3_rawpositions_high_5,minus3_rawpositions_high_5,
		    plus3_rawpositions_high_3,minus3_rawpositions_high_3,
#endif
		    plus3_rawpositions_5,minus3_rawpositions_5,
		    plus3_rawpositions_3,minus3_rawpositions_3,
		    plus3_nrawpositions_5,minus3_nrawpositions_5,
		    plus3_nrawpositions_3,minus3_nrawpositions_3);
	Stage1_fill_all_oligos(this3,queryuc_ptr_3,querylength3,genestrand);

	queryrc3 = (char *) MALLOC((querylength3+1)*sizeof(char));
	make_complement_buffered(queryrc3,queryuc_ptr_3,querylength3);

	debug(printf("LEVEL %d: RUNNING EXTENSION_SEARCH ON 3' READ\n",level));
	Extension_search(&found_score_3,&hits3_gplus,&hits3_gminus,
			 &queryfwd_plus_set_3,&queryfwd_minus_set_3,&queryrev_plus_set_3,&queryrev_minus_set_3,
			 this3,queryuc_ptr_3,queryrc3,querylength3,query3_compress_fwd,query3_compress_rev,
			 pairpool,dynprogL,dynprogM,dynprogR,oligoindices_minor,diagpool,cellpool,
			 nmismatches_allowed_3,genestrand,/*paired_end_p*/true,level);
      }
      
      /* Check for concordance */
      debug(printf("found scores: %d and %d (vs %d and %d allowed)\n",
		   found_score_5,found_score_3,nmismatches_allowed_5,nmismatches_allowed_3));
      debug(printf("hits5 plus: %d new.  hits5 minus: %d new.\n",List_length(hits5_gplus),List_length(hits5_gminus)));
      debug(printf("hits3 plus: %d new.  hits3 minus: %d new.\n",List_length(hits3_gplus),List_length(hits3_gminus)));
      hitpairs = Concordance_pair_up_genome(&(*abort_pairing_p),&adjacent_score,&concordant_score,hitpairs,
					    hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					    ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					    querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(2) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		   level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
    }

    /* Remaining: 3% of reads.  Segment_search by itself is too slow, but can constrain from the other end */
    if (*abort_pairing_p == false && adjacent_score > nmismatches_allowed) {
      if (found_score_5 >= found_score_3 && abort_exact5_p == false) {
	if (queryrc5 == NULL) {
	  queryrc5 = (char *) MALLOC((querylength5+1)*sizeof(char));
	  make_complement_buffered(queryrc5,queryuc_ptr_5,querylength5);
	}
	debug(printf("LEVEL %d: RUNNING KMER_SEARCH_GENOME_ENDS_APPROX ON 5' READ\n",level));
	Kmer_search_genome_ends_approx(&found_score_5,&hits5_gplus,&hits5_gminus,
#ifdef LARGE_GENOMES
				       plus5_rawpositions_high_5,minus5_rawpositions_high_5,
				       plus5_rawpositions_high_3,minus5_rawpositions_high_3,
#endif
				       plus5_rawpositions_5,plus5_nrawpositions_5,plus5_diagterms_5,
				       minus5_rawpositions_5,minus5_nrawpositions_5,minus5_diagterms_5,
				       plus5_rawpositions_3,plus5_nrawpositions_3,plus5_diagterms_3,
				       minus5_rawpositions_3,minus5_nrawpositions_3,minus5_diagterms_3,
				      
				       query5_compress_fwd,query5_compress_rev,querylength5,
				       genestrand,nmismatches_allowed_5,/*max_hits*/1000,
				       /*sizelimit*/3000,level);
      }

      if (found_score_3 >= found_score_5 && abort_exact3_p == false) {
	if (queryrc3 == NULL) {
	  queryrc3 = (char *) MALLOC((querylength3+1)*sizeof(char));
	  make_complement_buffered(queryrc3,queryuc_ptr_3,querylength3);
	}
	debug(printf("LEVEL %d: RUNNING KMER_SEARCH_GENOME_ENDS_APPROX ON 3' READ\n",level));
	Kmer_search_genome_ends_approx(&found_score_3,&hits3_gplus,&hits3_gminus,
#ifdef LARGE_GENOMES
				       plus3_rawpositions_high_5,minus3_rawpositions_high_5,
				       plus3_rawpositions_high_3,minus3_rawpositions_high_3,
#endif
				       plus3_rawpositions_5,plus3_nrawpositions_5,plus3_diagterms_5,
				       minus3_rawpositions_5,minus3_nrawpositions_5,minus3_diagterms_5,
				       plus3_rawpositions_3,plus3_nrawpositions_3,plus3_diagterms_3,
				       minus3_rawpositions_3,minus3_nrawpositions_3,minus3_diagterms_3,

				       query3_compress_fwd,query3_compress_rev,querylength3,
				       genestrand,nmismatches_allowed_3,/*max_hits*/1000,
				       /*sizelimit*/3000,level);
      }

      debug(printf("found scores: %d and %d (vs %d and %d allowed)\n",
		   found_score_5,found_score_3,nmismatches_allowed_5,nmismatches_allowed_3));
      debug(printf("hits5 plus: %d new.  hits5 minus: %d new.\n",List_length(hits5_gplus),List_length(hits5_gminus)));
      debug(printf("hits3 plus: %d new.  hits3 minus: %d new.\n",List_length(hits3_gplus),List_length(hits3_gminus)));
      hitpairs = Concordance_pair_up_genome(&(*abort_pairing_p),&adjacent_score,&concordant_score,hitpairs,
					    hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					    ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					    querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(3) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		   level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
    }


    if (*abort_pairing_p == false && adjacent_score > nmismatches_allowed) {
      debug(printf("SEGMENT SEARCH\n"));
      if (querylength5 < index1part) {
	/* Skip */
      } else {
	if (this5 == NULL) {
	  this5 = Stage1_new(querylength5);
	  Stage1_init(this5,querylength5,/*one5p*/false,
#ifdef LARGE_GENOMES
		      plus5_rawpositions_high_5,minus5_rawpositions_high_5,
		      plus5_rawpositions_high_3,minus5_rawpositions_high_3,
#endif
		      plus5_rawpositions_5,minus5_rawpositions_5,
		      plus5_rawpositions_3,minus5_rawpositions_3,
		      plus5_nrawpositions_5,minus5_nrawpositions_5,
		      plus5_nrawpositions_3,minus5_nrawpositions_3);
	  Stage1_fill_all_oligos(this5,queryuc_ptr_5,querylength5,genestrand);
	}
	Stage1_ptr_all_positions(this5,querylength5,genestrand);
	/* Need sizelimit to constrain segment search */
	sizelimit5 = determine_sizelimit(this5,querylength5);

	if (queryrc5 == NULL) {
	  queryrc5 = (char *) MALLOC((querylength5+1)*sizeof(char));
	  make_complement_buffered(queryrc5,queryuc_ptr_5,querylength5);
	}
      }

      if (querylength3 < index1part) {
	/* Skip */
      } else {
	if (this3 == NULL) {
	  this3 = Stage1_new(querylength3);
	  Stage1_init(this3,querylength3,/*one3p*/false,
#ifdef LARGE_GENOMES
		      plus3_rawpositions_high_5,minus3_rawpositions_high_5,
		      plus3_rawpositions_high_3,minus3_rawpositions_high_3,
#endif
		      plus3_rawpositions_5,minus3_rawpositions_5,
		      plus3_rawpositions_3,minus3_rawpositions_3,
		      plus3_nrawpositions_5,minus3_nrawpositions_5,
		      plus3_nrawpositions_3,minus3_nrawpositions_3);
	  Stage1_fill_all_oligos(this3,queryuc_ptr_3,querylength3,genestrand);
	}
	Stage1_ptr_all_positions(this3,querylength3,genestrand);
	/* Need sizelimit to constrain segment search */
	sizelimit3 = determine_sizelimit(this3,querylength3);
      
	if (queryrc3 == NULL) {
	  queryrc3 = (char *) MALLOC((querylength3+1)*sizeof(char));
	  make_complement_buffered(queryrc3,queryuc_ptr_3,querylength3);
	}
      }
    }

    /* TODO: Need branch if ladder5 and ladder3 are both empty */
    if (*abort_pairing_p == false && adjacent_score > nmismatches_allowed) {
      orig_found_score_5 = found_score_5;
      orig_found_score_3 = found_score_3;

      if (orig_found_score_5 >= orig_found_score_3 && querylength5 >= index1part) {
	debug(printf("LEVEL %d: RUNNING SEGMENT_SEARCH_GENOME ON 5' READ STARTING WITH SCORE %d\n",level,score_start_5));
	plus_records5 = Segment_identify(&plus_nrecords5,&plus_all_records5,
					 &plus_candidates5,&plus_ncandidates5,
#ifdef LARGE_GENOMES
					 this5->plus_positions_high,
#endif
					 this5->plus_positions,this5->plus_npositions,this5->plus_validp,
					 querylength5,sizelimit5,/*plusp*/true);
	minus_records5 = Segment_identify(&minus_nrecords5,&minus_all_records5,
					  &minus_candidates5,&minus_ncandidates5,
#ifdef LARGE_GENOMES
					  this5->minus_positions_high,
#endif
					  this5->minus_positions,this5->minus_npositions,this5->minus_validp,
					  querylength5,sizelimit5,/*plusp*/false);

	Segment_search_all(&found_score_5,&hits5_gplus,&hits5_gminus,
			   &plus_complete_paths_5,&minus_complete_paths_5,
			   
			   plus_records5,plus_nrecords5,plus_candidates5,plus_ncandidates5,
			   minus_records5,minus_nrecords5,minus_candidates5,minus_ncandidates5,
			   
			   queryuc_ptr_5,queryrc5,querylength5,
			   query5_compress_fwd,query5_compress_rev,
			   genestrand,/*paired_end_p*/true,level);
      }

      if (orig_found_score_3 >= orig_found_score_5 && querylength3 >= index1part) {
	debug(printf("LEVEL %d: RUNNING SEGMENT_SEARCH_GENOME ON 3' READ STARTING WITH SCORE %d\n",level,score_start_3));

	plus_records3 = Segment_identify(&plus_nrecords3,&plus_all_records3,
					 &plus_candidates3,&plus_ncandidates3,
#ifdef LARGE_GENOMES
					 this3->plus_positions_high,
#endif
					 this3->plus_positions,this3->plus_npositions,this3->plus_validp,
					 querylength3,sizelimit3,/*plusp*/true);
	minus_records3 = Segment_identify(&minus_nrecords3,&minus_all_records3,
					  &minus_candidates3,&minus_ncandidates3,
#ifdef LARGE_GENOMES
					  this3->minus_positions_high,
#endif
					  this3->minus_positions,this3->minus_npositions,this3->minus_validp,
					  querylength3,sizelimit3,/*plusp*/false);
	
	Segment_search_all(&found_score_3,&hits3_gplus,&hits3_gminus,
			   &plus_complete_paths_3,&minus_complete_paths_3,
			   
			   plus_records3,plus_nrecords3,plus_candidates3,plus_ncandidates3,
			   minus_records3,minus_nrecords3,minus_candidates3,minus_ncandidates3,
			   
			   queryuc_ptr_3,queryrc3,querylength3,
			   query3_compress_fwd,query3_compress_rev,
			   genestrand,/*paired_end_p*/true,level);
      }
      
      debug(printf("found scores: %d and %d (vs %d and %d allowed)\n",
		   found_score_5,found_score_3,nmismatches_allowed_5,nmismatches_allowed_3));
      debug(printf("hits5 plus: %d new.  hits5 minus: %d new.\n",List_length(hits5_gplus),List_length(hits5_gminus)));
      debug(printf("hits3 plus: %d new.  hits3 minus: %d new.\n",List_length(hits3_gplus),List_length(hits3_gminus)));
      hitpairs = Concordance_pair_up_genome(&(*abort_pairing_p),&adjacent_score,&concordant_score,hitpairs,
					    hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					    ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					    querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(4) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		   level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
      
      FREE(plus_candidates3);
      FREE(minus_candidates3);
      FREE(plus_candidates5);
      FREE(minus_candidates5);
    
      FREE(plus_records3);
      FREE(minus_records3);
      FREE(plus_records5);
      FREE(minus_records5);
    
      FREE(plus_all_records3);
      FREE(minus_all_records3);
      FREE(plus_all_records5);
      FREE(minus_all_records5);
    }

    debug(printf("COMPLETE PATHS: %d, %d, %d, %d\n",
		 List_length(plus_complete_paths_5),
		 List_length(minus_complete_paths_5),
		 List_length(plus_complete_paths_3),
		 List_length(minus_complete_paths_3)));

    /* Remaining: 0.2% of reads.  Generally few complete_paths, so try SEG_GMAP algorithm */

#if 0
    if (hitpairs == NULL) {
      debug(printf("RUNNING PATH_SOLVE_VIA_GMAP ON 5' READS (%d and %d)\n",
		   List_length(plus_complete_paths_5),List_length(minus_complete_paths_5)));
      Path_solve_via_gmap(&found_score_5,&hits5_gplus,&hits5_gminus,
			  plus_complete_paths_5,minus_complete_paths_5,genestrand,
			  queryuc_ptr_5,querylength5,query5_compress_fwd,query5_compress_rev,
			  pairpool,dynprogL,dynprogM,dynprogR,oligoindices_minor,diagpool,cellpool,
			  /*paired_end_p*/true,/*method*/SEGMENT_GMAP,level);
	  
      debug(printf("RUNNING PATH_SOLVE_VIA_GMAP ON 3' READS (%d and %d)\n",
		   List_length(plus_complete_paths_3),List_length(minus_complete_paths_3)));
      Path_solve_via_gmap(&found_score_3,&hits3_gplus,&hits3_gminus,
			  plus_complete_paths_3,minus_complete_paths_3,genestrand,
			  queryuc_ptr_3,querylength3,query3_compress_fwd,query3_compress_rev,
			  pairpool,dynprogL,dynprogM,dynprogR,oligoindices_minor,diagpool,cellpool,
			  /*paired_end_p*/true,/*method*/SEGMENT_GMAP,level);
      
      debug(printf("found scores: %d and %d (vs %d and %d allowed)\n",
		   found_score_5,found_score_3,nmismatches_allowed_5,nmismatches_allowed_3));
      debug(printf("hits5 plus: %d new.  hits5 minus: %d new.\n",List_length(hits5_gplus),List_length(hits5_gminus)));
      debug(printf("hits3 plus: %d new.  hits3 minus: %d new.\n",List_length(hits3_gplus),List_length(hits3_gminus)));
      hitpairs = Concordance_pair_up_genome(&(*abort_pairing_p),&adjacent_score,&concordant_score,hitpairs,
					    hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					    ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					    querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(5) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		   level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
    }
#endif

    Univdiag_list_gc(&plus_complete_paths_5);
    Univdiag_list_gc(&minus_complete_paths_5);
    Univdiag_list_gc(&plus_complete_paths_3);
    Univdiag_list_gc(&minus_complete_paths_3);

    Stage1_free(&this5);
    Stage1_free(&this3);


    /* Remap to transcriptome */
    /* TODO: Need to iterate through Ladder_T object */



    /* Find distant splicing.  Use concordant_score, rather than
       adjacent_score as criterion, since distant splicing cannot find
       adjacent pairs */
    if (*abort_pairing_p == false && concordant_score > nmismatches_allowed) {
      
      /* distant splicing on 5' read */
      if (found_score_5 >= found_score_3) {
	if ((done_level_5 = found_score_5 + subopt_levels) > nmismatches_allowed_5) {
	  done_level_5 = nmismatches_allowed_5;
	}
	debug(printf("done_level_5 %d = found_score_5 %d + subopt_levels %d\n",
		     done_level_5,found_score_5,subopt_levels));
	
	/* min of max(trim5,trim3) over each hit */      
	min_trim_5 = Ladder_minimax_trim(ladder5_plus,ladder5_minus,querylength5);

	debug(printf("For 5' end, comparing min_trim %d with min_distantsplicing_end_matches %d\n",
		     min_trim_5,min_distantsplicing_end_matches));
	if (min_trim_5 < min_distantsplicing_end_matches) {
	  /* Don't find distant splicing */
	  debug(printf("For 5' end, skipping distant splicing because min_trim %d < min_distantsplicing_end_matches %d\n",
		       min_trim_5,min_distantsplicing_end_matches));
	
	} else if ((max_splice_mismatches = done_level_5 - distantsplicing_penalty) >= 0) {
	  if (novelsplicingp == true) {
	    debug(printf("For 5' end, candidate for distant splicing because min_trim %d >= min_distantsplicing_end_matches %d\n",
			 min_trim_5,min_distantsplicing_end_matches));
	    debug(printf("Have sets: %d %d %d %d\n",
			 List_length(queryfwd_plus_set_5),List_length(queryfwd_minus_set_5),
			 List_length(queryrev_plus_set_5),List_length(queryrev_minus_set_5)));
	    Distant_rna_solve(&found_score_5,&hits5_gplus,&hits5_gminus,
			      
			      queryfwd_plus_set_5,queryfwd_minus_set_5,
			      queryrev_plus_set_5,queryrev_minus_set_5,
			      
			      query5_compress_fwd,query5_compress_rev,querylength5,
			      /*user_maxlevel*/done_level_5,max_splice_mismatches,
			      genestrand,/*first_read_p*/true,level);
	  } else {
	    /* TODO: Implement distant DNA fusions */
	  }
	}
      }

      /* Distant splicing on 3' read */
      if (found_score_3 >= found_score_5) {
	if ((done_level_3 = found_score_3 + subopt_levels) > nmismatches_allowed_3) {
	  done_level_3 = nmismatches_allowed_3;
	}
	debug(printf("done_level_3 %d = found_score_3 %d + subopt_levels %d\n",
		     done_level_3,found_score_3,subopt_levels));

	/* min of max(trim5,trim3) over each hit */      
	min_trim_3 = Ladder_minimax_trim(ladder3_plus,ladder3_minus,querylength3);
	
	debug(printf("For 3' end, comparing min_trim %d with min_distantsplicing_end_matches %d\n",
		     min_trim_3,min_distantsplicing_end_matches));
	if (min_trim_3 < min_distantsplicing_end_matches) {
	  /* Don't find distant splicing */
	  debug(printf("For 3'end, skipping distant splicing because min_trim %d < min_distantsplicing_end_matches %d\n",
		       min_trim_3,min_distantsplicing_end_matches));
	
	} else if ((max_splice_mismatches = done_level_3 - distantsplicing_penalty) >= 0) {
	  if (novelsplicingp == true) {
	    debug(printf("For 3' end, candidate for distant splicing because min_trim %d >= min_distantsplicing_end_matches %d\n",
			 min_trim_3,min_distantsplicing_end_matches));
	    debug(printf("Have sets: %d %d %d %d\n",
			 List_length(queryfwd_plus_set_3),List_length(queryfwd_minus_set_3),
			 List_length(queryrev_plus_set_3),List_length(queryrev_minus_set_3)));
	    Distant_rna_solve(&found_score_3,&hits3_gplus,&hits3_gminus,
			      
			      queryfwd_plus_set_3,queryfwd_minus_set_3,
			      queryrev_plus_set_3,queryrev_minus_set_3,
			      
			      query3_compress_fwd,query3_compress_rev,querylength3,
			      /*user_maxlevel*/done_level_3,max_splice_mismatches,
			      genestrand,/*first_read_p*/false,level);
	  } else {
	    /* TODO: Implement distant DNA */
	  }
	}
      }

      debug(printf("found scores: %d and %d (vs %d and %d allowed)\n",
		   found_score_5,found_score_3,nmismatches_allowed_5,nmismatches_allowed_3));
      debug(printf("hits5 plus: %d new.  hits5 minus: %d new.\n",List_length(hits5_gplus),List_length(hits5_gminus)));
      debug(printf("hits3 plus: %d new.  hits3 minus: %d new.\n",List_length(hits3_gplus),List_length(hits3_gminus)));
      hitpairs = Concordance_pair_up_distant(&(*abort_pairing_p),&concordant_score,&(*samechr),&(*conc_transloc),hitpairs,
					     hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					     ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					     querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(6) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		     level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
    }


    /* Terminals */
    if (*abort_pairing_p == false && adjacent_score > nmismatches_allowed) {
      /* Need to recompute, because of new distant hits or because distant splicing was not run */
      min_trim_5 = Ladder_minimax_trim(ladder5_plus,ladder5_minus,querylength5);

      if (min_trim_5 < min_distantsplicing_end_matches) {
	debug(printf("For 5' end, skipping terminals because min_trim %d < min_distantsplicing_end_matches %d\n",
		     min_trim_5,min_distantsplicing_end_matches));
      } else {
	max_terminal_mismatches = done_level_5;
	if ((hits5_gplus = Terminal_solve_plus(&found_score_5,
					       queryfwd_plus_set_5,queryrev_plus_set_5,
					       query5_compress_fwd,querylength5,
					       /*user_maxlevel*/done_level_5,max_terminal_mismatches,
					       genestrand,/*first_read_p*/true,level)) == NULL) {
	  debug(printf("Got no terminals\n"));
	} else {
	  debug(printf("Got %d terminals\n",List_length(hits5_gplus)));
	}

	if ((hits5_gminus = Terminal_solve_minus(&found_score_5,
						 queryfwd_minus_set_5,queryrev_minus_set_5,
						 query5_compress_rev,querylength5,
						 /*user_maxlevel*/done_level_5,max_terminal_mismatches,
						 genestrand,/*first_read_p*/true,level)) == NULL) {
	  debug(printf("Got no terminals\n"));
	} else {
	  debug(printf("Got %d terminals\n",List_length(hits5_gminus)));
	}
      }


      /* Need to recompute, because of new distant hits or because distant splicing was not run */
      min_trim_3 = Ladder_minimax_trim(ladder3_plus,ladder3_minus,querylength3);

      if (min_trim_3 < min_distantsplicing_end_matches) {
	debug(printf("For 3'end, skipping terminals because min_trim %d < min_distantsplicing_end_matches %d\n",
		     min_trim_3,min_distantsplicing_end_matches));
      } else {
	max_terminal_mismatches = done_level_3;
	if ((hits3_gplus = Terminal_solve_plus(&found_score_3,
					       queryfwd_plus_set_3,queryrev_plus_set_3,
					       query3_compress_fwd,querylength3,
					       /*user_maxlevel*/done_level_3,max_terminal_mismatches,
					       genestrand,/*first_read_p*/false,level)) == NULL) {
	  debug(printf("Got no terminals\n"));
	} else {
	  debug(printf("Got %d terminals\n",List_length(hits3_gplus)));
	}

	if ((hits3_gminus = Terminal_solve_minus(&found_score_3,
						 queryfwd_minus_set_3,queryrev_minus_set_3,
						 query3_compress_rev,querylength3,
						 /*user_maxlevel*/done_level_3,max_terminal_mismatches,
						 genestrand,/*first_read_p*/false,level)) == NULL) {
	  debug(printf("Got no terminals\n"));
	} else {
	  debug(printf("Got %d terminals\n",List_length(hits3_gminus)));
	}
      }

      hitpairs = Concordance_pair_up_genome(&(*abort_pairing_p),&adjacent_score,&concordant_score,hitpairs,
					    hits5_gplus,hits5_gminus,hits3_gplus,hits3_gminus,
					    ladder5_plus,ladder5_minus,ladder3_plus,ladder3_minus,
					    querylength5,querylength3,maxpairedpaths,genestrand); level++;
      List_free(&hits5_gplus); List_free(&hits5_gminus); List_free(&hits3_gplus); List_free(&hits3_gminus);
      debug(printf("(7) After level %d, we have adjacent_score %d, concordant_score %d (vs allowed %d), %d pairs\n",
		     level,adjacent_score,concordant_score,nmismatches_allowed,List_length(hitpairs)));
    }

    Elt_gc(&queryfwd_plus_set_3);
    Elt_gc(&queryfwd_minus_set_3);
    Elt_gc(&queryrev_plus_set_3);
    Elt_gc(&queryrev_minus_set_3);
    Elt_gc(&queryfwd_plus_set_5);
    Elt_gc(&queryfwd_minus_set_5);
    Elt_gc(&queryrev_plus_set_5);
    Elt_gc(&queryrev_minus_set_5);

    FREE(queryrc3);
    FREE(queryrc5);
  }

  *hits5 = Ladder_hitlist(ladder5_plus,NULL);
  *hits5 = Ladder_hitlist(ladder5_minus,*hits5);

  *hits3 = Ladder_hitlist(ladder3_plus,NULL);
  *hits3 = Ladder_hitlist(ladder3_minus,*hits3);

  Ladder_gc_duplicates(ladder5_plus);
  Ladder_gc_duplicates(ladder5_minus);
  Ladder_gc_duplicates(ladder3_plus);
  Ladder_gc_duplicates(ladder3_minus);

  Ladder_free(&ladder5_plus);
  Ladder_free(&ladder5_minus);
  Ladder_free(&ladder3_plus);
  Ladder_free(&ladder3_minus);

  return hitpairs;
}


/* Have three lists: hitpairs, samechr, and conc_transloc => result */
static Stage3pair_T *
consolidate_paired_results (int *npaths_primary, int *npaths_altloc, int *first_absmq, int *second_absmq, Pairtype_T *final_pairtype,
			    Stage3end_T **stage3array5, int *nhits5_primary, int *nhits5_altloc, int *first_absmq5, int *second_absmq5,
			    Stage3end_T **stage3array3, int *nhits3_primary, int *nhits3_altloc, int *first_absmq3, int *second_absmq3,
			    List_T hitpairs, List_T samechr, List_T conc_transloc, List_T hits5, List_T hits3,

			    char *queryuc_ptr_5, int querylength5, char *queryuc_ptr_3, int querylength3,
			    char *quality_string_5, char *quality_string_3,
			    Compress_T query5_compress_fwd, Compress_T query5_compress_rev,
			    Compress_T query3_compress_fwd, Compress_T query3_compress_rev,

			    int min_coverage_5, int min_coverage_3, Oligoindex_array_T oligoindices_minor,
			    Pairpool_T pairpool, Diagpool_T diagpool, Cellpool_T cellpool,
			    Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR) {
  Stage3pair_T *stage3pairarray, stage3pair, newpair;
  Stage3end_T hit5, hit3;
  List_T result, singlehits5, singlehits3, p;
  int genestrand;
  Pairtype_T pairtype;
  int best_nmatches_paired, best_nmatches_paired_5, best_nmatches_paired_3;

  
  *final_pairtype = choose_among_paired(&best_nmatches_paired,&best_nmatches_paired_5,&best_nmatches_paired_3,
					hitpairs,samechr,conc_transloc);
  debug16(printf("Entered consolidate_paired_results with final_pairtype %d\n",*final_pairtype));


  if (*final_pairtype == CONCORDANT) {
    /* Have concordant results */
    debug16(printf("Have %d concordant results\n",List_length(hitpairs)));
    for (p = samechr; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&samechr);

    for (p = conc_transloc; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&conc_transloc);
	  
    if (1) {
      debug16(printf("Via transcriptome: Before removing overlaps, %d results\n",List_length(hitpairs)));
      result = Stage3pair_optimal_score(hitpairs,querylength5,querylength3,/*finalp*/true);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/true);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);
      
      result = Stage3pair_resolve_insides(result,queryuc_ptr_5,queryuc_ptr_3,
					  query5_compress_fwd,query5_compress_rev,
					  query3_compress_fwd,query3_compress_rev,
					  pairpool,dynprogL,dynprogM,dynprogR,
					  oligoindices_minor,diagpool,cellpool);

      result = Stage3pair_resolve_multimapping(result);
      /* result = Stage3pair_sort_distance(result); */
      debug16(printf("After removing overlaps, %d results\n",List_length(result)));

    } else if (gmap_improvement_p == false) {
      debug16(printf("No GMAP improvement: Before removing overlaps, %d results\n",List_length(hitpairs)));
      result = Stage3pair_optimal_score(hitpairs,querylength5,querylength3,/*finalp*/true);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/true);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);

      result = Stage3pair_resolve_insides(result,queryuc_ptr_5,queryuc_ptr_3,
					  query5_compress_fwd,query5_compress_rev,
					  query3_compress_fwd,query3_compress_rev,
					  pairpool,dynprogL,dynprogM,dynprogR,
					  oligoindices_minor,diagpool,cellpool);

      result = Stage3pair_resolve_multimapping(result);
      /* result = Stage3pair_sort_distance(result); */
      debug16(printf("After removing overlaps, %d results\n",List_length(result)));

    } else {
      debug16(printf("GMAP improvement: Before removing overlaps, %d results\n",List_length(hitpairs)));
      result = Stage3pair_optimal_score(hitpairs,querylength5,querylength3,/*finalp*/false);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/false);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/false);
      result = Stage3pair_resolve_multimapping(result);
      /* result = Stage3pair_sort_distance(result); */
      debug16(printf("After removing overlaps, %d results\n",List_length(result)));

#ifdef TODO
      result = align_pair_with_gmap(&(*final_pairtype),result,
				    queryuc_ptr_5,queryuc_ptr_3,querylength5,querylength3,
				    oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR,
				    cutoff_level_5,cutoff_level_3,/*expect_concordant_p*/true);
      if (Stage3pair_sense_consistent_p(result) == false) {
	debug16(printf("sense is not consistent\n"));
	result = align_pair_with_gmap(&(*final_pairtype),result,
				      queryuc_ptr_5,queryuc_ptr_3,querylength5,querylength3,
				      oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR,
				      cutoff_level_5,cutoff_level_3,/*expect_concordant_p*/true);
      }
#endif

      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/true);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);

      result = Stage3pair_resolve_insides(result,queryuc_ptr_5,queryuc_ptr_3,
					  query5_compress_fwd,query5_compress_rev,
					  query3_compress_fwd,query3_compress_rev,
					  pairpool,dynprogL,dynprogM,dynprogR,
					  oligoindices_minor,diagpool,cellpool);

      result = Stage3pair_resolve_multimapping(result);
    }

  } else if (*final_pairtype == PAIRED_UNSPECIFIED) {
    /* Have paired results */
    debug16(printf("Have paired unspecified\n"));
    for (p = hitpairs; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&hitpairs);

    for (p = conc_transloc; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&conc_transloc);

    if (gmap_improvement_p == false) {
      debug16(printf("No GMAP improvement: Before removing overlaps, %d results\n",List_length(samechr)));
      result = Stage3pair_optimal_score(samechr,querylength5,querylength3,/*finalp*/true);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/true);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);
      result = Stage3pair_resolve_multimapping(result);
    } else {
      debug16(printf("GMAP improvement: Before removing overlaps, %d results\n",List_length(samechr)));
      result = Stage3pair_optimal_score(samechr,querylength5,querylength3,/*finalp*/false);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/false);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/false);
      result = Stage3pair_resolve_multimapping(result);

#ifdef TODO
      result = align_pair_with_gmap(&(*final_pairtype),result,
				    queryuc_ptr_5,queryuc_ptr_3,querylength5,querylength3,
				    oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR,
				    cutoff_level_5,cutoff_level_3,/*expect_concordant_p*/false);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);
      result = Stage3pair_remove_overlaps(result,/*translocp*/false,/*finalp*/true);
      result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);
#endif

      if (Stage3pair_concordantp(result) == true) {
	debug16(printf("Found remaining concordant solution, so removing non-concordant ones\n"));
	*final_pairtype = CONCORDANT;
	result = Stage3pair_filter_nonconcordant(result);

	result = Stage3pair_resolve_insides(result,queryuc_ptr_5,queryuc_ptr_3,
					    query5_compress_fwd,query5_compress_rev,
					    query3_compress_fwd,query3_compress_rev,
					    pairpool,dynprogL,dynprogM,dynprogR,
					    oligoindices_minor,diagpool,cellpool);

	debug16(printf("Concordant results: %d\n",List_length(result)));
      } else {
	*final_pairtype = PAIRED_UNSPECIFIED;
      }

      result = Stage3pair_resolve_multimapping(result);
    }

    *final_pairtype = CONCORDANT; /* Because paired terminals are each concordant */


  } else if (*final_pairtype == CONCORDANT_TRANSLOCATIONS) {
    debug16(printf("Have %d concordant translocation results\n",List_length(conc_transloc)));
    for (p = hitpairs; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&hitpairs);

    for (p = samechr; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&samechr);

    result = Stage3pair_optimal_score(conc_transloc,querylength5,querylength3,/*finalp*/true);
    result = Stage3pair_remove_overlaps(result,/*translocp*/true,/*finalp*/true);
    result = Stage3pair_optimal_score(result,querylength5,querylength3,/*finalp*/true);

    result = Stage3pair_resolve_insides(result,queryuc_ptr_5,queryuc_ptr_3,
					query5_compress_fwd,query5_compress_rev,
					query3_compress_fwd,query3_compress_rev,
					pairpool,dynprogL,dynprogM,dynprogR,
					oligoindices_minor,diagpool,cellpool);

    result = Stage3pair_resolve_multimapping(result);
    debug16(printf("Finally, have %d concordant translocation results\n",List_length(result)));

  } else {
    debug16(printf("Have unpaired results\n"));
    /* Need to free conc_transloc, since we can get here with multiple results */
    for (p = conc_transloc; p != NULL; p = List_next(p)) {
      stage3pair = (Stage3pair_T) List_head(p);
      Stage3pair_free(&stage3pair);
    }
    List_free(&conc_transloc);

    result = (List_T) NULL;
  }


  if (result == NULL) {
    singlehits5 = Stage3end_optimal_score(hits5,querylength5,/*finalp*/true);
    /* singlehits5 = Stage3end_reject_trimlengths(singlehits5); */
    singlehits5 = Stage3end_linearize_5(singlehits5);
    singlehits5 = Stage3end_remove_overlaps(singlehits5,/*finalp*/true);
    singlehits5 = Stage3end_optimal_score(singlehits5,querylength5,/*finalp*/true);
    singlehits5 = Stage3end_resolve_multimapping(singlehits5);

    singlehits3 = Stage3end_optimal_score(hits3,querylength3,/*finalp*/true);
    /* singlehits3 = Stage3end_reject_trimlengths(singlehits3); */
    singlehits3 = Stage3end_linearize_3(singlehits3);
    singlehits3 = Stage3end_remove_overlaps(singlehits3,/*finalp*/true);
    singlehits3 = Stage3end_optimal_score(singlehits3,querylength3,/*finalp*/true);
    singlehits3 = Stage3end_resolve_multimapping(singlehits3);

    debug16(printf("5' end has %d hits and 3' end has %d hits\n",
		   List_length(singlehits5),List_length(singlehits3)));

    if (List_length(singlehits5) == 1 && List_length(singlehits3) == 1) {
      hit5 = (Stage3end_T) List_head(singlehits5);
      hit3 = (Stage3end_T) List_head(singlehits3);
      if ((genestrand = Stage3end_genestrand(hit5)) == Stage3end_genestrand(hit3) &&
	  (pairtype = Stage3_determine_pairtype(hit5,hit3,/*stage3pair*/NULL)) != UNPAIRED) {
	/* Convert unpaired uniq to a paired uniq */
	debug16(printf("Converting unpaired uniq to paired uniq, with initial pairtype %s\n",Pairtype_string(pairtype)));
	if ((newpair = Stage3pair_new(hit5,hit3,genestrand,pairtype,
				      /*expect_concordant_p*/pairtype == CONCORDANT ? true : false,
				      /*transcriptome_guided_p*/false)) != NULL) {
	  result = List_push(NULL,(void *) newpair);
	  
	  *nhits5_primary = *nhits5_altloc = 0;
	  *nhits3_primary = *nhits3_altloc = 0;
	  *stage3array5 = *stage3array3 = (Stage3end_T *) NULL;
	  
	  if (Stage3pair_altlocp(newpair) == true) {
	    *npaths_primary = 0;
	    *npaths_altloc = 1;
	  } else {
	    *npaths_primary = 1;
	    *npaths_altloc = 0;
	  }
	  if (pairtype == CONCORDANT) {
	    debug16(printf("final pairtype is CONCORDANT\n"));
	    *final_pairtype = CONCORDANT;
	    
	    result = Stage3pair_resolve_insides(result,queryuc_ptr_5,queryuc_ptr_3,
						query5_compress_fwd,query5_compress_rev,
						query3_compress_fwd,query3_compress_rev,
						pairpool,dynprogL,dynprogM,dynprogR,
						oligoindices_minor,diagpool,cellpool);
	    
	  } else {
	    debug16(printf("final pairtype is PAIRED_UNSPECIFIED\n"));
	    *final_pairtype = PAIRED_UNSPECIFIED;
	  }

	  stage3pairarray = (Stage3pair_T *) CALLOC_OUT(1,sizeof(Stage3pair_T));
	  stage3pairarray[0] = (Stage3pair_T) List_head(result);
	  List_free(&result);
	  
#if 0
	  Stage3pair_privatize(stage3pairarray,/*npairs*/1);
#endif
	  Stage3pair_eval_and_sort(/*npaths*/(*npaths_primary) + (*npaths_altloc),
				   &(*first_absmq),&(*second_absmq),stage3pairarray,
				   queryuc_ptr_5,queryuc_ptr_3,quality_string_5,quality_string_3);
	  
	  stage3list_gc(&singlehits3);
	  stage3list_gc(&singlehits5);
	  debug16(printf("1 Exiting consolidate_paired_results\n"));
	  return stage3pairarray;
	}
      }
    }

    /* Fall through: halfmapping or unpaired */
    *npaths_primary = *npaths_altloc = 0;
    *final_pairtype = UNPAIRED;
	  
    singlehits5 = Stage3end_filter_coverage(singlehits5,min_coverage_5);
    if (singlehits5 == NULL) {
      *nhits5_primary = *nhits5_altloc = 0;
      *stage3array5 = (Stage3end_T *) NULL;
    } else {
#if 0
      singlehits5 = Stage3end_unalias_circular(singlehits5);
#else
      singlehits5 = Stage3end_remove_circular_alias(singlehits5); /* Contains a call to unalias_circular */
      singlehits5 = Stage3end_remove_duplicates(singlehits5); /* Aliases can cause duplicates */
      Stage3end_count_hits(&(*nhits5_primary),&(*nhits5_altloc),singlehits5);
#endif
      *stage3array5 = (Stage3end_T *) List_to_array_out(singlehits5,NULL); List_free(&singlehits5); /* Return value */
    }

    singlehits3 = Stage3end_filter_coverage(singlehits3,min_coverage_3);
    if (singlehits3 == NULL) {
      *nhits3_primary = *nhits3_altloc = 0;
      *stage3array3 = (Stage3end_T *) NULL;
    } else {
#if 0
      singlehits3 = Stage3end_unalias_circular(singlehits3);
#else
      singlehits3 = Stage3end_remove_circular_alias(singlehits3); /* Contains a call to unalias_circular */
      singlehits3 = Stage3end_remove_duplicates(singlehits3); /* Aliases can cause duplicates */
      Stage3end_count_hits(&(*nhits3_primary),&(*nhits3_altloc),singlehits3);
#endif
      *stage3array3 = (Stage3end_T *) List_to_array_out(singlehits3,NULL); List_free(&singlehits3); /* Return value */
    }

    if ((*nhits5_primary) + (*nhits5_altloc) > 0) {
      if ((*nhits3_primary) + (*nhits3_altloc) == 1) {
      /* Use single 3' hit to guide sorting of multiple 5' hits */
        *stage3array5 = Stage3end_eval_and_sort_guided((*nhits5_primary) + (*nhits5_altloc),
   	                                               &(*first_absmq5),&(*second_absmq5),/*guide*/(*stage3array3)[0],
						       *stage3array5,queryuc_ptr_5,quality_string_5,/*displayp*/true);
      } else {
        *stage3array5 = Stage3end_eval_and_sort((*nhits5_primary) + (*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
 						*stage3array5,queryuc_ptr_5,quality_string_5,/*displayp*/true);
      }
    }

    if ((*nhits3_primary) + (*nhits3_altloc) > 0) {
      if ((*nhits5_primary) + (*nhits5_altloc) == 1) {
	/* Use single 5' hit to guide sorting of multiple 3' hits */
        *stage3array3 = Stage3end_eval_and_sort_guided((*nhits3_primary) + (*nhits3_altloc),
                                                       &(*first_absmq3),&(*second_absmq3),/*guide*/(*stage3array5)[0],
						       *stage3array3,queryuc_ptr_3,quality_string_3,/*displayp*/true);
      } else {
        *stage3array3 = Stage3end_eval_and_sort((*nhits3_primary) + (*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
						*stage3array3,queryuc_ptr_3,quality_string_3,/*displayp*/true);
      }
    }
    debug16(printf("Result is NULL, and we have %d hits on 5' end and %d hits on 3' end\n",
		   (*nhits5_primary) + (*nhits5_altloc),(*nhits3_primary) + (*nhits3_altloc)));
    debug16(printf("2 Exiting consolidate_paired_results\n"));
    return (Stage3pair_T *) NULL;

  } else {
    debug16(printf("final pairtype is %d\n",*final_pairtype));
    debug16(printf("Result is not NULL (%d paths), and we fall through to concordant, paired, or transloc pairs\n",
		   List_length(result)));

    result = Stage3pair_filter_coverage(result,min_coverage_5,min_coverage_3);
    if (result == NULL) {
      *npaths_primary = *npaths_altloc = 0;
      stage3list_gc(&hits3);
      stage3list_gc(&hits5);
      *nhits5_primary = *nhits5_altloc = 0;
      *nhits3_primary = *nhits3_altloc = 0;
      *stage3array5 = *stage3array3 = (Stage3end_T *) NULL;
      debug16(printf("3 Exiting consolidate_paired_results\n"));
      return (Stage3pair_T *) NULL;

    } else {
      /* result != NULL */
      /* Concordant, paired, or transloc pairs found.  Remove single hits. */
      Stage3pair_count_hits(&(*npaths_primary),&(*npaths_altloc),result);
      stage3pairarray = (Stage3pair_T *) List_to_array_out(result,NULL); List_free(&result); /* Return value */
#if 0
      Stage3pair_privatize(stage3pairarray,/*npaths*/(*npaths_primary) + (*npaths_altloc));
#endif
      Stage3pair_eval_and_sort(/*npaths*/(*npaths_primary) + (*npaths_altloc),
			       &(*first_absmq),&(*second_absmq),
			       stage3pairarray,queryuc_ptr_5,queryuc_ptr_3,quality_string_5,quality_string_3);
      stage3list_gc(&hits3);
      stage3list_gc(&hits5);

      *nhits5_primary = *nhits5_altloc = 0;
      *nhits3_primary = *nhits3_altloc = 0;
      *stage3array5 = *stage3array3 = (Stage3end_T *) NULL;
      debug16(printf("4 Exiting consolidate_paired_results\n"));

      return stage3pairarray;
    }
  }
}


Stage3pair_T *
Stage1_paired_read (int *npaths_primary, int *npaths_altloc, int *first_absmq, int *second_absmq, Pairtype_T *final_pairtype,
		    Stage3end_T **stage3array5, int *nhits5_primary, int *nhits5_altloc, int *first_absmq5, int *second_absmq5,
		    Stage3end_T **stage3array3, int *nhits3_primary, int *nhits3_altloc, int *first_absmq3, int *second_absmq3,
		    Shortread_T queryseq5, Shortread_T queryseq3, Oligoindex_array_T oligoindices_minor,
		    Pairpool_T pairpool, Diagpool_T diagpool, Cellpool_T cellpool,
		    Dynprog_T dynprogL, Dynprog_T dynprogM, Dynprog_T dynprogR) {
  Stage3pair_T *stage3pairarray;
  bool abort_pairing_p, geneplus_abort_pairing_p, geneminus_abort_pairing_p;
  List_T hitpairs, geneplus_hitpairs, geneminus_hitpairs;
  List_T hits5, hits3, geneplus_hits5, geneplus_hits3, geneminus_hits5, geneminus_hits3;
  List_T samechr, geneplus_samechr, geneminus_samechr;
  List_T conc_transloc, geneplus_conc_transloc, geneminus_conc_transloc;

  int querylength5, querylength3;
  int min_coverage_5, min_coverage_3;
  char *queryuc_ptr_5, *queryuc_ptr_3, *quality_string_5, *quality_string_3;
  Compress_T query5_compress_fwd, query5_compress_rev, query3_compress_fwd, query3_compress_rev;


  querylength5 = Shortread_fulllength(queryseq5);
  querylength3 = Shortread_fulllength(queryseq3);

  if (user_mincoverage_float < 0.0) {
    min_coverage_5 = min_coverage_3 = 0;
  } else if (user_mincoverage_float > 0.0 && user_mincoverage_float < 1.0) {
    min_coverage_5 = (int) rint(user_mincoverage_float * (double) querylength5);
    min_coverage_3 = (int) rint(user_mincoverage_float * (double) querylength3);
  } else {
    min_coverage_5 = min_coverage_3 = (int) user_mincoverage_float;
  }

  queryuc_ptr_5 = Shortread_fullpointer_uc(queryseq5);
  queryuc_ptr_3 = Shortread_fullpointer_uc(queryseq3);
  quality_string_5 = Shortread_quality_string(queryseq5);
  quality_string_3 = Shortread_quality_string(queryseq3);

  query5_compress_fwd = Compress_new_fwd(queryuc_ptr_5,querylength5);
  query5_compress_rev = Compress_new_rev(queryuc_ptr_5,querylength5);
  query3_compress_fwd = Compress_new_fwd(queryuc_ptr_3,querylength3);
  query3_compress_rev = Compress_new_rev(queryuc_ptr_3,querylength3);


  if (mode == STANDARD || mode == CMET_STRANDED || mode == ATOI_STRANDED || mode == TTOC_STRANDED) {
    hitpairs = paired_read(&abort_pairing_p,&hits5,&hits3,&samechr,&conc_transloc,
			   queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
			   query5_compress_fwd,query5_compress_rev,
			   query3_compress_fwd,query3_compress_rev,
			   /*genestrand*/0,oligoindices_minor,
			   pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);
    if (abort_pairing_p == true) {
      /* Too many results.  Realign separately */
      debug(printf("abort_pairing_p is true\n"));
      paired_results_free(hitpairs,samechr,conc_transloc,hits5,hits3,querylength5,querylength3);

      *stage3array5 = Stage1_single_read(&(*nhits5_primary),&(*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
					 queryseq5,oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);
      *stage3array3 = Stage1_single_read(&(*nhits3_primary),&(*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
					 queryseq3,oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

      *npaths_primary = *npaths_altloc = 0;
      *final_pairtype = UNPAIRED;
      stage3pairarray = (Stage3pair_T *) NULL;

    } else {
      *final_pairtype = CONCORDANT;
      stage3pairarray =
	consolidate_paired_results(&(*npaths_primary),&(*npaths_altloc),&(*first_absmq),&(*second_absmq),&(*final_pairtype),
				   &(*stage3array5),&(*nhits5_primary),&(*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
				   &(*stage3array3),&(*nhits3_primary),&(*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
				   hitpairs,samechr,conc_transloc,hits5,hits3,

				   queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				   quality_string_5,quality_string_3,
				   query5_compress_fwd,query5_compress_rev,query3_compress_fwd,query3_compress_rev,

				   min_coverage_5,min_coverage_3,
				   oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);
    }

  } else if (mode == CMET_NONSTRANDED || mode == ATOI_NONSTRANDED || mode == TTOC_NONSTRANDED) {
    geneplus_hitpairs = paired_read(&geneplus_abort_pairing_p,&geneplus_hits5,&geneplus_hits3,
				    &geneplus_samechr,&geneplus_conc_transloc,
				    queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				    query5_compress_fwd,query5_compress_rev,
				    query3_compress_fwd,query3_compress_rev,
				    /*genestrand*/+1,oligoindices_minor,
				    pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

    geneminus_hitpairs = paired_read(&geneminus_abort_pairing_p,&geneminus_hits5,&geneminus_hits3,
				     &geneminus_samechr,&geneminus_conc_transloc,
				     queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				     query5_compress_fwd,query5_compress_rev,
				     query3_compress_fwd,query3_compress_rev,
				     /*genestrand*/+2,oligoindices_minor,
				     pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

    if (geneplus_abort_pairing_p == true && geneminus_abort_pairing_p == true) {
      /* Too many results.  Realign separately */
      debug(printf("abort_pairing_p is true\n"));
      paired_results_free(geneplus_hitpairs,geneplus_samechr,geneplus_conc_transloc,
			  geneplus_hits5,geneplus_hits3,querylength5,querylength3);
      paired_results_free(geneminus_hitpairs,geneminus_samechr,geneminus_conc_transloc,
			  geneminus_hits5,geneminus_hits3,querylength5,querylength3);

      *stage3array5 = Stage1_single_read(&(*nhits5_primary),&(*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
					 queryseq5,oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);
      *stage3array3 = Stage1_single_read(&(*nhits3_primary),&(*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
					 queryseq3,oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

      *npaths_primary = *npaths_altloc = 0;
      *final_pairtype = UNPAIRED;

    } else if (geneplus_abort_pairing_p == true) {
      paired_results_free(geneplus_hitpairs,geneplus_samechr,geneplus_conc_transloc,
			  geneplus_hits5,geneplus_hits3,querylength5,querylength3);

      *final_pairtype = CONCORDANT;
      stage3pairarray =
	consolidate_paired_results(&(*npaths_primary),&(*npaths_altloc),&(*first_absmq),&(*second_absmq),&(*final_pairtype),
				   &(*stage3array5),&(*nhits5_primary),&(*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
				   &(*stage3array3),&(*nhits3_primary),&(*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
				   geneminus_hitpairs,geneminus_samechr,geneminus_conc_transloc,
				   geneminus_hits5,geneminus_hits3,

				   queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				   quality_string_5,quality_string_3,
				   query5_compress_fwd,query5_compress_rev,query3_compress_fwd,query3_compress_rev,

				   min_coverage_5,min_coverage_3,
				   oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);

    } else if (geneminus_abort_pairing_p == true) {
      paired_results_free(geneminus_hitpairs,geneminus_samechr,geneminus_conc_transloc,
			  geneminus_hits5,geneminus_hits3,querylength5,querylength3);

      *final_pairtype = CONCORDANT;
      stage3pairarray =
	consolidate_paired_results(&(*npaths_primary),&(*npaths_altloc),&(*first_absmq),&(*second_absmq),&(*final_pairtype),
				   &(*stage3array5),&(*nhits5_primary),&(*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
				   &(*stage3array3),&(*nhits3_primary),&(*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
				   geneplus_hitpairs,geneplus_samechr,geneplus_conc_transloc,
				   geneplus_hits5,geneplus_hits3,

				   queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				   quality_string_5,quality_string_3,
				   query5_compress_fwd,query5_compress_rev,query3_compress_fwd,query3_compress_rev,

				   min_coverage_5,min_coverage_3,
				   oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);
    } else {
      hits5 = List_append(geneplus_hits5,geneminus_hits5);
      hits3 = List_append(geneplus_hits3,geneminus_hits3);
      hitpairs = List_append(geneplus_hitpairs,geneminus_hitpairs);
      samechr = List_append(geneplus_samechr,geneminus_samechr);
      conc_transloc = List_append(geneplus_conc_transloc,geneminus_conc_transloc);

      stage3pairarray =
	consolidate_paired_results(&(*npaths_primary),&(*npaths_altloc),&(*first_absmq),&(*second_absmq),&(*final_pairtype),
				   &(*stage3array5),&(*nhits5_primary),&(*nhits5_altloc),&(*first_absmq5),&(*second_absmq5),
				   &(*stage3array3),&(*nhits3_primary),&(*nhits3_altloc),&(*first_absmq3),&(*second_absmq3),
				   hitpairs,samechr,conc_transloc,hits5,hits3,

				   queryuc_ptr_5,querylength5,queryuc_ptr_3,querylength3,
				   quality_string_5,quality_string_3,
				   query5_compress_fwd,query5_compress_rev,query3_compress_fwd,query3_compress_rev,

				   min_coverage_5,min_coverage_3,
				   oligoindices_minor,pairpool,diagpool,cellpool,dynprogL,dynprogM,dynprogR);
    }

  } else {
    fprintf(stderr,"Do not recognize mode %d\n",mode);
    abort();
  }

  Compress_free(&query5_compress_fwd);
  Compress_free(&query5_compress_rev);
  Compress_free(&query3_compress_fwd);
  Compress_free(&query3_compress_rev);

  return stage3pairarray;
}


void
Stage1hr_setup (Univ_IIT_T transcript_iit_in, Transcriptome_T transcriptome_in, Genome_T transcriptomebits_in,
		bool use_only_transcriptome_p_in,

		Indexdb_T indexdb_fwd_in, Indexdb_T indexdb_rev_in, Indexdb_T indexdb_tr_in, Localdb_T localdb_in,

		int index1part_tr_in, int index1part_in, int index1interval_in, 
		double user_maxlevel_float_in, double user_mincoverage_float_in,

		Univ_IIT_T chromosome_iit_in, int nchromosomes_in,
		Genome_T genomecomp_in, Genome_T genomebits_in, Genome_T genomebits_alt_in,
		Mode_T mode_in, int maxpaths_search_in, int maxpaths_report_in,

		bool find_dna_chimeras_p_in, bool distances_observed_p_in, int subopt_levels_in,
		int max_middle_insertions, int max_middle_deletions,

		bool novelsplicingp_in,
		Chrpos_T shortsplicedist_in, Chrpos_T shortsplicedist_known_in, Chrpos_T shortsplicedist_novelend_in,
		Chrpos_T min_intronlength_in, Chrpos_T expected_pairlength_in, Chrpos_T pairlength_deviation_in,

		int distantsplicing_penalty_in,
		int min_distantsplicing_end_matches_in, int min_distantsplicing_identity_in,

		int nullgap_in, int maxpeelback_in, int maxpeelback_distalmedial_in,
		int extramaterial_end_in, int extramaterial_paired_in,
		int gmap_mode_in, int trigger_score_for_gmap_in, int gmap_allowance_in,
		int max_gmap_pairsearch_in, int max_gmap_segments_in,
		int max_gmap_improvement_in) {
  bool gmapp = false;

  use_only_transcriptome_p = use_only_transcriptome_p_in;

  transcript_iit = transcript_iit_in;
  transcriptome = transcriptome_in;
  transcriptomebits = transcriptomebits_in;

  indexdb_fwd = indexdb_fwd_in;
  indexdb_rev = indexdb_rev_in;
  indexdb_tr = indexdb_tr_in;
  localdb = localdb_in;

  index1part_tr = index1part_tr_in;
  index1part = index1part_in;
  index1interval = index1interval_in;

  user_maxlevel_float = user_maxlevel_float_in;
  user_mincoverage_float = user_mincoverage_float_in;

  chromosome_iit = chromosome_iit_in;
  circular_typeint = Univ_IIT_typeint(chromosome_iit,"circular");
  nchromosomes = nchromosomes_in;

  genomecomp = genomecomp_in;
  genomebits = genomebits_in;
  genomebits_alt = genomebits_alt_in;

#ifdef HAVE_64_BIT
  leftreadshift = 64 - index1part - index1part;
  oligobase_mask = ~(~ (Oligospace_T) 0 << 2*index1part);
#else
  leftreadshift = 32 - index1part - index1part;
  oligobase_mask = ~(~ (Oligospace_T) 0 << 2*index1part);
#endif

  poly_a = POLY_A & oligobase_mask;
  poly_t = POLY_T & oligobase_mask;

  mode = mode_in;
  maxpaths_search = maxpaths_search_in;
  maxpaths_report = maxpaths_report_in;

  find_dna_chimeras_p = find_dna_chimeras_p_in;
  distances_observed_p = distances_observed_p_in;

  subopt_levels = subopt_levels_in;

  novelsplicingp = novelsplicingp_in;
  shortsplicedist = shortsplicedist_in;

  overall_max_distance = shortsplicedist;
  if (max_middle_deletions > (int) overall_max_distance) {
    overall_max_distance = max_middle_deletions;
  }
  if (max_middle_insertions > (int) overall_max_distance) {
    overall_max_distance = max_middle_insertions;
  }

  min_intronlength = min_intronlength_in;
  expected_pairlength = expected_pairlength_in;
  pairlength_deviation = pairlength_deviation_in;

  distantsplicing_penalty = distantsplicing_penalty_in;
  min_distantsplicing_end_matches = min_distantsplicing_end_matches_in;
  min_distantsplicing_identity = min_distantsplicing_identity_in;

  nullgap = nullgap_in;
  maxpeelback = maxpeelback_in;
  maxpeelback_distalmedial = maxpeelback_distalmedial_in;
  extramaterial_end = extramaterial_end_in;
  extramaterial_paired = extramaterial_paired_in;

  gmap_mode = gmap_mode_in;

  gmap_segments_p = false;
  gmap_pairsearch_p = false;
  gmap_improvement_p = false;

  fprintf(stderr,"GMAP modes:");
  if ((gmap_mode & GMAP_PAIRSEARCH) != 0) {
    if (gmapp == true) {
      fprintf(stderr,",");
    } else {
      gmapp = true;
    }
    fprintf(stderr," pairsearch");
    gmap_pairsearch_p = true;
  }
  if ((gmap_mode & GMAP_ENDS) != 0) {
    if (gmapp == true) {
      fprintf(stderr,",");
    } else {
      gmapp = true;
    }
    fprintf(stderr," ends");
    gmap_segments_p = true;
  }
  if ((gmap_mode & GMAP_IMPROVEMENT) != 0) {
    if (gmapp == true) {
      fprintf(stderr,",");
    } else {
      gmapp = true;
    }
    fprintf(stderr," improvement");
    gmap_improvement_p = true;
  }
  if (gmapp == false) {
    fprintf(stderr," none");
  }
  fprintf(stderr,"\n");


  trigger_score_for_gmap = trigger_score_for_gmap_in;
  gmap_allowance = gmap_allowance_in;

  max_gmap_pairsearch = max_gmap_pairsearch_in;
  max_gmap_segments = max_gmap_segments_in;
  max_gmap_improvement = max_gmap_improvement_in;

  if (genomebits_alt_in != NULL) {
    snpp = true;
  } else {
    snpp = false;
  }

  return;
}
